####################################################################
#                                                                  #
#    ensemble_md,                                                  #
#    a python package for running GROMACS simulation ensembles     #
#                                                                  #
#    Written by Wei-Tse Hsu <wehs7661@colorado.edu>                #
#    Copyright (c) 2022 University of Colorado Boulder             #
#                                                                  #
####################################################################
"""
The :obj:`.analyze_matrix` module provides methods for analyzing matrices obtained from REXEE.
"""
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib import cm
from ensemble_md.utils.exceptions import ParseError
from ensemble_md.utils.exceptions import ParameterError


def calc_transmtx(log_file, expanded_ensemble=True):
    """
    Parses the log file to get the transition matrix of an expanded ensemble
    or replica exchange simulation. Notably, a theoretical transition matrix
    is only available in expanded ensemble.

    Parameters
    ----------
    log_file : str
        The log file to be parsed.
    expanded_ensemble : bool
        Whether the simulation is expanded ensemble or replica exchange

    Returns
    -------
    empirical : np.ndarray
        The final empirical state transition matrix.
    theoretical : None or np.ndarray
        The final theoretical state transition matrix.
    diff_matrix : None or np.ndarray
        The difference between the theortial and empirical state transition matrix (empirical - theoretical).
    """
    f = open(log_file, "r")
    lines = f.readlines()
    f.close()

    lines.reverse()

    n = -1
    theoretical_found, empirical_found = False, False
    for l in lines:  # noqa: E741
        n += 1
        if "Empirical Transition Matrix" in l:  # This will be found first
            empirical_found = True
            n_states = int(lines[n - 1].split()[-1])
            empirical = np.zeros([n_states, n_states])
            for i in range(n_states):
                if expanded_ensemble is True:
                    empirical[i] = [float(k) for k in lines[n - 2 - i].split()[:-1]]
                else:    # replica exchange
                    empirical[i] = [float(k) for k in lines[n - 2 - i].split()[1:-1]]

        if "Transition Matrix" in l and "Empirical" not in l:    # only occurs in expanded ensemble
            theoretical_found = True
            theoretical = np.zeros([n_states, n_states])
            for i in range(n_states):
                theoretical[i] = [float(k) for k in lines[n - 2 - i].split()[:-1]]

        if theoretical_found is True and empirical_found is True:
            diff_matrix = empirical - theoretical
            break

    if theoretical_found is False:
        theoretical = None
        diff_matrix = None

        if empirical_found is False:
            raise ParseError(f"No transition matrices found in {log_file}.")

    return empirical, theoretical, diff_matrix


def calc_equil_prob(trans_mtx):
    """
    Calculates the equilibrium probability of each state from the state transition matrix.
    The input state transition matrix can be either left or right stochastic, although the left
    stochastic ones are not common in GROMACS. Generally, transition matrices in GROMACS are either
    doubly stochastic (replica exchange), or right stochastic (expanded ensemble). For the latter case,
    the staionary distribution vector is the left eigenvector corresponding to the eigenvalue 1
    of the transition matrix. (For the former case, it's either left or right eigenvector corresponding
    to the eigenvalue 1 - as the left and right eigenvectors are the same for a doubly stochasti matrix.)

    Parameters
    ----------
    trans_mtx : np.ndarray
        The input state transition matrix

    Returns
    -------
    equil_prob : np.ndarray
    """
    check_row = sum([np.isclose(np.sum(trans_mtx[i]), 1) for i in range(len(trans_mtx))])
    check_col = sum([np.isclose(np.sum(trans_mtx[:, i]), 1) for i in range(len(trans_mtx))])

    if check_row == len(trans_mtx):  # note that this also include doubly stochastic matrices
        # Right or doubly stachstic matrices - calculate the left eigenvector
        eig_vals, eig_vecs = np.linalg.eig(trans_mtx.T)
    elif check_col == len(trans_mtx):
        # Left stochastic matrix - calcualte the right eigenvector
        eig_vals, eig_vecs = np.linalg.eig(trans_mtx)
    else:
        result_str = "The input transition matrix is neither right nor left stochastic, so the equilibrium probability cannot be calculated."  # noqa: E501
        result_str += "This might be a result of poor sampling. Check your state-space trajectory to troubleshoot."
        print(result_str)
        return None

    # The eigenvector corresponding to the eigenvalue eig_vals[i] is eig_vecs[:, i]
    close_to_1_idx = np.isclose(eig_vals, 1, atol=1e-4)
    equil_prob = eig_vecs[:, close_to_1_idx]  # note that this is normalized
    equil_prob /= np.sum(equil_prob)   # So the sum of all components is 1
    equil_prob = np.array([i.real for i in equil_prob if i.imag == 0])  # get rid of the imaginary part if it is 0
    return equil_prob


def calc_spectral_gap(trans_mtx, atol=1e-8):
    """
    Calculates the spectral gap of the input transition matrix.

    Parameters
    ----------
    trans_mtx : np.ndarray
        The input state transition matrix
    atol: float
        The absolute tolerance for checking the sum of columns and rows.

    Returns
    -------
    spectral_gap : float
        The spectral gap of the input transitio n matrix
    eig_vals : list
        The list of eigenvalues
    """
    check_row = sum([np.isclose(np.sum(trans_mtx[i]), 1, atol=atol) for i in range(len(trans_mtx))])
    check_col = sum([np.isclose(np.sum(trans_mtx[:, i]), 1, atol=atol) for i in range(len(trans_mtx))])

    if check_row == len(trans_mtx):
        eig_vals, eig_vecs = np.linalg.eig(trans_mtx.T)
    elif check_col == len(trans_mtx):
        eig_vals, eig_vecs = np.linalg.eig(trans_mtx)
    else:
        result_str = "The input transition matrix is neither right nor left stochastic, so the spectral gap cannot be calculated."  # noqa: E501
        result_str += "This might be a result of poor sampling. Check your state-space trajectory to troubleshoot."
        print(result_str)
        return None

    eig_vals = np.sort(eig_vals)[::-1]  # descending order
    if np.isclose(eig_vals[0], 1, atol=1e-4) is False:
        raise ParameterError(f'The largest eigenvalue of the input transition matrix {eig_vals[0]} is not close to 1.')

    spectral_gap = np.abs(eig_vals[0]) - np.abs(eig_vals[1])

    return spectral_gap, eig_vals


def split_transmtx(trans_mtx, n_sim, n_sub):
    """
    Split the input transition matrix into blocks of smaller matrices corresponding to
    difrerent alchemical ranges of different replicas. Notably, the function assumes
    homogeneous shifts and number of states across replicas. Also, the blocks of the
    transition matrix is generally not doubly stochastic but right stochastic even if
    the input is doubly stochastic.

    Parameters
    ----------
    trans_mtx : np.ndarray
        The input state transition matrix to split
    n_sim : int
        The number of replicas in REXEE.
    n_sub : int
        The number of states for each replica.

    Returns
    -------
    sub_mtx: list
        Blocks of transition matrices split from the input.
    """
    sub_mtx = []
    ranges = [[i, i + n_sub] for i in range(n_sim)]   # A list of lists containing the min/max of alchemcial ranges
    for i in range(len(ranges)):
        bounds = ranges[i]
        # Below with numpy.copy, trans_mtx will be indirectly changed by the next line
        mtx = np.copy(trans_mtx[bounds[0]:bounds[1]][:, bounds[0]:bounds[1]])
        mtx /= np.sum(mtx, axis=1)[:, None]    # Divide each row by the sum of the row to rescale probabilities
        sub_mtx.append(mtx)

    return sub_mtx


def plot_matrix(matrix, png_name, title=None, start_idx=0):
    """
    Visualizes a matrix a in heatmap.

    Parameters
    ----------
    matrix : np.ndarray
        The matrix to be visualized
    png_name : str
        The file name of the output PNG file (including the extension).
    title : str
        The title of the plot
    start_idx : int
        The starting value of the state index
    """

    sns.set_context(
        rc={"family": "sans-serif",
            "sans-serif": ["DejaVu Sans"],
            "size": 5})

    K = len(matrix)
    plt.figure(figsize=(K / 1.5, K / 1.5))     # or figsize=(K / 1.5, K / 1.5)
    annot_matrix = np.zeros([K, K])  # matrix for annotating values

    mask = []
    for i in range(K):
        mask.append([])
        for j in range(len(matrix[0])):
            if matrix[i][j] < 0.005:
                mask[-1].append(True)
            else:
                mask[-1].append(False)

    for i in range(K):
        for j in range(K):
            annot_matrix[i, j] = round(matrix[i, j], 2)

    x_tick_labels = y_tick_labels = np.arange(start_idx, start_idx + K)
    ax = sns.heatmap(
        matrix,
        cmap="YlGnBu",
        linecolor="silver",
        linewidth=0.25,
        annot=annot_matrix,
        square=True,
        mask=matrix < 0.005,
        fmt=".2f",
        cbar=False,
        xticklabels=x_tick_labels,
        yticklabels=y_tick_labels,
    )
    ax.xaxis.tick_top()
    ax.tick_params(length=0)
    cmap = cm.get_cmap("YlGnBu")  # to get the facecolor
    ax.set_facecolor(cmap(0))  # use the brightest color (value = 0)
    for _, spine in ax.spines.items():
        spine.set_visible(True)  # add frames to the heat map
    plt.annotate("$\lambda$", xy=(0, 0), xytext=(-0.45, -0.20))  # noqa: W605
    if title is None:
        pass
    else:
        plt.title(title, fontsize=10, weight="bold")
    plt.tight_layout(pad=1.0)

    plt.savefig(png_name, dpi=600)
    plt.close()



####################################################################
#                                                                  #
#    ensemble_md,                                                  #
#    a python package for running GROMACS simulation ensembles     #
#                                                                  #
#    Written by Wei-Tse Hsu <wehs7661@colorado.edu>                #
#    Copyright (c) 2022 University of Colorado Boulder             #
#                                                                  #
####################################################################
"""
The :obj:`.replica_exchange_EE` module provides functions for setting up and
replica exchange and expanded ensemble (REXEE) simulations.
"""
import os
import sys
import copy
import yaml
import random
import warnings
import importlib
import subprocess
import numpy as np
from mpi4py import MPI
from itertools import combinations
from collections import OrderedDict
from alchemlyb.parsing.gmx import _get_headers as get_headers
from alchemlyb.parsing.gmx import _extract_dataframe as extract_dataframe

import ensemble_md
from ensemble_md.utils import utils
from ensemble_md.utils import gmx_parser
from ensemble_md.utils.exceptions import ParameterError

comm = MPI.COMM_WORLD
rank = comm.Get_rank()


class ReplicaExchangeEE:
    """
    This class provides a variety of functions useful for setting up and running
    replica exchange (REX) of expanded ensemble (EE), or REXEE simulations.
    Upon instantiation, all parameters in the YAML file will be assigned to an
    attribute in the class. In addition to these variables, below is a list of
    attributes of the class. (All the the attributes are assigned by :obj:`set_params`
    unless otherwise noted.)

    :ivar gmx_path: The absolute path of the GROMACS exectuable.
    :ivar gmx_version: The version of the GROMACS executable.
    :ivar yaml: The input YAML file used to instantiate the class. Assigned by the :code:`__init__` function.
    :ivar warnings: Warnings about parameter specification in either YAML or MDP files.
    :ivar reformatted_mdp: Whether the templated MDP file has been reformatted by replacing hyphens
        with underscores or not.
    :ivar template: The instance of the :obj:`MDP` class based on the template MDP file.
    :ivar dt: The simulation timestep in ps.
    :ivar temp: The simulation temperature in Kelvin.
    :ivar fixed_weights: Whether the weights will be fixed during the simulation (according to the template MDP file).
    :ivar updating_weights: The list of weights as a function of time (since the last update of the Wang-Landau
        incrementor) for different replicas. The length is equal to the number of replicas. This is only relevant for
        weight-updating simulations.
    :ivar equilibrated_weights: The equilibrated weights of different replicas. For weight-equilibrating simulations,
        this list is initialized as a list of empty lists. Otherwise (weight-fixed), it is initialized as a list of
        :code:`None`.
    :ivar current_wl_delta: The current value of the Wang-Landau incrementor. This is only relevent for weight-updating
        simulations.
    :ivar kT: 1 kT in kJ/mol at the simulation temperature.
    :ivar lambda_types: The types of lambda variables involved in expanded ensemble simulations, e.g.
        :code:`fep_lambdas`, :code:`mass_lambdas`, :code:`coul_lambdas`, etc.
    :ivar n_tot: The total number of states for all replicas.
    :ivar n_sub: The numbmer of states for each replica. The current implementation assumes homogenous replicas.
    :ivar state_ranges: A list of list of state indices for each replica.
    :ivar equil: A list of times it took to equilibrated the weights for different replicas. This
        list is initialized with a list of -1, where -1 means that the weights haven't been equilibrated. Also,
        a value of 0 means that the simulation is a fixed-weight simulation.
    :ivar n_rejected: The number of proposed exchanges that have been rejected. Updated by :obj:`.accept_or_reject`.
    :ivar n_swap_attempts: The number of swaps attempted so far. This does not include the cases
        where there is no swappable pair. Updated by :obj:`.get_swapping_pattern`.
    :ivar n_emtpy_swappable: The number of times when there was no swappable pair.
    :ivar rep_trajs: The replica-space trajectories of all configurations.
    :ivar configs: A list that thows the current configuration index that each replica is sampling.
    :ivar g_vecs: The time series of the (processed) whole-range alchemical weights. If no weight combination is
        applied, this list will just be a list of :code:`None`'s.
    :ivar df_data_type: The type of data (either :math:`u_{nk}` or :math:`dH/dλ`) that will be used for
        free energy calculations if :code:`df_method` is :code:`True`.
    :ivar modify_coords_fn: The function (callable) in the external module (specified as :code:`modify_coords` in
        the input YAML file) for modifying coordinates at exchanges.
    """

    def __init__(self, yaml_file, analysis=False):
        self.yaml = yaml_file
        self.set_params(analysis)

    def set_params(self, analysis):
        """
        Sets up or reads in the user-defined parameters from a yaml file and an MDP template.
        This function is called to instantiate the class in the :code:`__init__` function of
        class. Specifically, it does the following:

          1. Sets up constants.
          2. Reads in parameters from a YAML file.
          3. Handles YAML parameters.
          4. Checks if the parameters in the YAML file are well-defined.
          5. Reformats the input MDP file to replace all hyphens with underscores.
          6. Reads in parameters from the MDP template.

        After instantiation, the class instance will have attributes corresponding to
        each of the parameters specified in the YAML file. For a full list of the parameters that
        can be specified in the YAML file, please refer to :ref:`doc_parameters`.

        :param yaml_file: The file name of the YAML file for specifying the parameters for REXEE.
        :type yaml_file: str
        :param analysis: Whether the instantiation of the class is for data analysis of REXEE simulations.
            The default is :code:`False`
        :type analysis: bool

        :raises ParameterError:

              - If a required parameter is not specified in the YAML file.
              - If a specified parameter is not recognizable.
              - If a specified option is not available for a parameter.
              - If the data type or range (e.g., positive or negative) of a parameter is not correct.
              - If an invalid MDP file is detected.
        """
        self.warnings = []  # Store warnings, if any.

        # Step 0: Set up constants
        k = 1.380649e-23
        NA = 6.0221408e23

        # Step 1: Read in parameters from the YAML file.
        with open(self.yaml) as f:
            params = yaml.load(f, Loader=yaml.FullLoader)

        for attr in params:
            setattr(self, attr, params[attr])

        # Step 2: Handle the compulsory YAML parameters
        required_args = [
            "gmx_executable",
            "gro",
            "top",
            "mdp",
            "n_sim",
            "n_iter",
            "s",
        ]
        for i in required_args:
            if hasattr(self, i) is False or getattr(self, i) is None:
                raise ParameterError(
                    f"Required parameter '{i}' not specified in {self.yaml}."
                )  # noqa: F405

        # Step 3: Handle the optional YAML parameters
        # Key: Optional argument; Value: Default value
        optional_args = {
            "add_swappables": None,
            "modify_coords": None,
            "nst_sim": None,
            "proposal": 'exhaustive',
            "w_combine": False,
            "w_mean_type": 'simple',
            "N_cutoff": 1000,
            "hist_corr": False,
            "verbose": True,
            "mdp_args": None,
            "grompp_args": None,
            "runtime_args": None,
            "n_ckpt": 100,
            "rm_cpt": True,
            "msm": False,
            "free_energy": False,
            "subsampling_avg": False,
            "df_spacing": 1,
            "df_ref": None,
            "df_method": "MBAR",
            "err_method": "propagate",
            "n_bootstrap": 50,
            "seed": None,
            # "n_ex": 'N^3',   # only active for multiple swaps.
        }
        for i in optional_args:
            if hasattr(self, i) is False or getattr(self, i) is None:
                setattr(self, i, optional_args[i])

        # all_args: Arguments that can be specified in the YAML file.
        all_args = required_args + list(optional_args.keys())
        for i in params:
            if i not in all_args:
                self.warnings.append(f'Warning: Parameter "{i}" specified in the input YAML file is not recognizable.')

        # Step 4: Check if the parameters in the YAML file are well-defined
        if self.proposal not in [None, 'single', 'neighboring', 'exhaustive']:  # deprecated option: multiple
            raise ParameterError("The specified proposal scheme is not available. Available options include 'single', 'neighboring', and 'exhaustive'.")  # noqa: E501

        if self.df_method not in [None, 'TI', 'BAR', 'MBAR']:
            raise ParameterError("The specified free energy estimator is not available. Available options include 'TI', 'BAR', and 'MBAR'.")  # noqa: E501

        if self.err_method not in [None, 'propagate', 'bootstrap']:
            raise ParameterError("The specified method for error estimation is not available. Available options include 'propagate', and 'bootstrap'.")  # noqa: E501

        params_int = ['n_sim', 'n_iter', 's', 'N_cutoff', 'df_spacing', 'n_ckpt', 'n_bootstrap']  # integer parameters  # noqa: E501
        if self.nst_sim is not None:
            params_int.append('nst_sim')
        """
        if self.n_ex != 'N^3':  # no need to add "and self.proposal == 'multiple' since if multiple swaps are not used, n_ex=1"  # noqa: E501
            params_int.append('n_ex')
        """
        if self.seed is not None:
            params_int.append('seed')
        for i in params_int:
            if type(getattr(self, i)) != int:
                raise ParameterError(f"The parameter '{i}' should be an integer.")

        params_pos = ['n_sim', 'n_iter', 'n_ckpt', 'df_spacing', 'n_bootstrap']  # positive parameters
        if self.nst_sim is not None:
            params_pos.append('nst_sim')
        for i in params_pos:
            if getattr(self, i) <= 0:
                raise ParameterError(f"The parameter '{i}' should be positive.")

        """
        if self.n_ex != 'N^3' and self.n_ex < 0:
            raise ParameterError("The parameter 'n_ex' should be non-negative.")
        """

        if self.s < 0:
            raise ParameterError("The parameter 's' should be non-negative.")

        if self.N_cutoff < 0 and self.N_cutoff != -1:
            raise ParameterError("The parameter 'N_cutoff' should be non-negative unless no weight correction is needed, i.e. N_cutoff = -1.")  # noqa: E501

        params_str = ['gro', 'top', 'mdp', 'gmx_executable']
        # First check if self.gro and self.top are lists and check their lengths
        check_files = ['gro', 'top']  # just for checking file types that can take multiple inputs
        for i in check_files:
            if isinstance(getattr(self, i), list):
                params_str.remove(i)
                if len(getattr(self, i)) != self.n_sim:
                    raise ParameterError(f"The number of the input {i.upper()} files must be the same as the number of replicas, if multiple are specified.")  # noqa: E501
        if self.modify_coords is not None:
            params_str.append('modify_coords')
        for i in params_str:
            if type(getattr(self, i)) != str:
                raise ParameterError(f"The parameter '{i}' should be a string.")

        params_bool = ['verbose', 'rm_cpt', 'msm', 'free_energy', 'subsampling_avg', 'w_combine']
        for i in params_bool:
            if type(getattr(self, i)) != bool:
                raise ParameterError(f"The parameter '{i}' should be a boolean variable.")

        params_list = ['add_swappables', 'df_ref']
        for i in params_list:
            if getattr(self, i) is not None and not isinstance(getattr(self, i), list):
                raise ParameterError(f"The parameter '{i}' should be a list.")

        params_dict = ['mdp_args', 'grompp_args', 'runtime_args']
        for i in params_dict:
            if getattr(self, i) is not None and not isinstance(getattr(self, i), dict):
                raise ParameterError(f"The parameter '{i}' should be a dictionary.")

        if self.add_swappables is not None:
            if not isinstance(self.add_swappables, list):
                raise ParameterError("The parameter 'add_swappables' should be a nested list.")
            for sublist in self.add_swappables:
                if not isinstance(sublist, list):
                    raise ParameterError("The parameter 'add_swappables' should be a nested list.")
                for item in sublist:
                    if not isinstance(item, int) or item < 0:
                        raise ParameterError("Each number specified in 'add_swappables' should be a non-negative integer.")  # noqa: E501

        if self.mdp_args is not None:
            for key in self.mdp_args.keys():
                if not isinstance(key, str):
                    raise ParameterError("All keys specified in 'mdp_args' should be strings.")
                else:
                    if '-' in key:
                        raise ParameterError("Parameters specified in 'mdp_args' must use underscores in place of hyphens.")  # noqa: E501
            for val_list in self.mdp_args.values():
                if len(val_list) != self.n_sim:
                    raise ParameterError("The number of values specified for each key in 'mdp_args' should be the same as the number of replicas.")  # noqa: E501

        # Step 5: Reformat the input MDP file to replace all hypens with underscores.
        self.reformatted_mdp = ReplicaExchangeEE.reformat_MDP(self.mdp)

        # Step 6: Read in parameters from the MDP template
        self.template = gmx_parser.MDP(self.mdp)
        self.dt = self.template["dt"]  # ps
        self.temp = self.template["ref_t"]

        if self.nst_sim is None:
            self.nst_sim = self.template["nsteps"]

        if 'wl_scale' in self.template.keys():
            if self.template['wl_scale'] == []:
                # If wl_scale in the MDP file is a blank (i.e. fixed weights), mdp['wl_scale'] will be an empty list.
                # This is the only case where mdp['wl_scale'] is a numpy array.
                self.fixed_weights = True
                self.equilibrated_weights = [None for i in range(self.n_sim)]
            else:
                self.fixed_weights = False
                self.equilibrated_weights = [[] for i in range(self.n_sim)]
                self.updating_weights = [[] for i in range(self.n_sim)]
                self.current_wl_delta = [0 for i in range(self.n_sim)]
        else:
            self.fixed_weights = True
            self.equilibrated_weights = [None for i in range(self.n_sim)]

        if self.fixed_weights is True:
            if self.N_cutoff != -1 or self.w_combine is True:
                self.warnings.append('Warning: The weight correction/weight combination method is specified but will not be used since the weights are fixed.')  # noqa: E501
                # In the case that the warning is ignored, enforce the defaults.
                self.N_cutoff = -1
                self.w_combine = False

        if 'lmc_seed' in self.template and self.template['lmc_seed'] != -1:
            self.warnings.append('Warning: We recommend setting lmc_seed as -1 so the random seed is different for each iteration.')  # noqa: E501

        if 'gen_seed' in self.template and self.template['gen_seed'] != -1:
            self.warnings.append('Warning: We recommend setting gen_seed as -1 so the random seed is different for each iteration.')  # noqa: E501

        if 'gen_vel' not in self.template or ('gen_vel' in self.template and self.template['gen_vel'] == 'no'):
            self.warnings.append('Warning: We recommend generating new velocities for each iteration to avoid potential issues with detailed balance.')  # noqa: E501

        if self.nst_sim % self.template['nstlog'] != 0:
            raise ParameterError(
                'The parameter "nstlog" must be a factor of the parameter "nst_sim" specified in the YAML file.')

        if self.nst_sim % self.template['nstdhdl'] != 0:
            raise ParameterError(
                'The parameter "nstdhdl" must be a factor of the parameter "nst_sim" specified in the YAML file.')

        if self.template['nstexpanded'] % self.template['nstdhdl'] != 0:
            raise ParameterError(
                'In REXEE, the parameter "nstdhdl" must be a factor of the parameter "nstexpanded", or the calculation of acceptance ratios might be wrong.')  # noqa: E501

        if self.mdp_args is not None:
            if 'lmc_seed' in self.mdp_args and -1 not in self.mdp_args['lmc_seed']:
                self.warnings.append('Warning: We recommend setting lmc_seed as -1 so the random seed is different for each iteration.')  # noqa: E501

            if 'gen_seed' in self.mdp_args and -1 not in self.mdp_args['gen_seed']:
                self.warnings.append('Warning: We recommend setting gen_seed as -1 so the random seed is different for each iteration.')  # noqa: E501

            if 'gen_vel' in self.mdp_args and 'no' in self.mdp_args['gen_vel']:
                self.warnings.append('Warning: We recommend generating new velocities for each iteration to avoid potential issues with the detailed balance.')  # noqa: E501

            if 'nstlog' in self.mdp_args and sum(self.nst_sim % np.array(self.mdp_args['nstlog'])) != 0:
                raise ParameterError(
                    'The parameter "nstlog" must be a factor of the parameter "nst_sim" specified in the YAML file.')

            if 'nstdhdl' in self.mdp_args and sum(self.nst_sim % np.array(self.mdp_args['nstdhdl'])) != 0:
                raise ParameterError(
                    'The parameter "nstdhdl" must be a factor of the parameter "nst_sim" specified in the YAML file.')

            if 'nstexpanded' in self.mdp_args and 'nstdhdl' in self.mdp_args and sum(np.array(self.mdp_args['nstexpanded']) % np.array(self.mdp_args['nstdhdl'])) != 0:  # noqa: E501
                raise ParameterError(
                    'In REXEE, the parameter "nstdhdl" must be a factor of the parameter "nstexpanded", or the calculation of acceptance ratios might be wrong.')  # noqa: E501

        if 'pull' in self.template and self.template['pull'] == 'yes':
            pull_ncoords = self.template['pull_ncoords']
            self.set_ref_dist = []
            for i in range(pull_ncoords):
                if self.template[f'pull_coord{i+1}_geometry'] == 'distance':
                    if self.template[f'pull_coord{i+1}_start'] == 'yes':
                        self.set_ref_dist.append(True)  # starting from the second iteration, set pull_coord*_init.
                        if 'pull_nstxout' not in self.template:
                            self.warnings.append('A non-zero value should be specified for pull_nstxout if pull_coord*_start is set to yes.')  # noqa: E501
                        if self.template['pull_nstxout'] == 0:
                            self.warnings.append('A non-zero value should be specified for pull_nstxout if pull_coord*_start is set to yes.')  # noqa: E501
                    else:
                        self.set_ref_dist.append(False)  # Here we assume that the user know what reference distance to use.  # noqa: E501
                else:
                    self.set_ref_dist.append(False)  # we only deal with distance restraints for now.

        # Step 7: Set up derived parameters
        # 7-1. kT in kJ/mol
        self.kT = k * NA * self.temp / 1000  # 1 kT in kJ/mol

        # 7-2. Figure out what types of lambda variables are involved
        # Here is we possible lambda types in the order read by GROMACS, which is likely also the order when being printed to the log file.  # noqa: E501
        # See https://gitlab.com/gromacs/gromacs/-/blob/main/src/gromacs/gmxpreprocess/readir.cpp#L2543
        lambdas_types_all = ['fep_lambdas', 'mass_lambdas', 'coul_lambdas', 'vdw_lambdas', 'bonded_lambdas', 'restraint_lambdas', 'temperature_lambdas']  # noqa: E501
        self.lambda_types = []  # lambdas specified in the MDP file
        for i in lambdas_types_all:
            if i in self.template.keys():  # there shouldn't be parameters like "fep-lambdas" after reformatting the MDP file  # noqa: E501
                self.lambda_types.append(i)

        # 7-3. Total # of states: n_tot = n_sub * n_sim - (n_overlap) * (n_sim - 1), where n_overlap = n_sub - s
        self.n_tot = len(self.template[self.lambda_types[0]])

        # 7-4. Number of states of each replica (assuming the same for all rep)
        self.n_sub = self.n_tot - self.s * (self.n_sim - 1)
        if self.n_sub < 1:
            raise ParameterError(
                f"There must be at least two states for each replica (current value: {self.n_sub}). The current specified configuration (n_tot={self.n_tot}, n_sim={self.n_sim}, s={self.s}) does not work for REXEE.")  # noqa: E501

        # 7-5. A list of sets of state indices
        start_idx = [i * self.s for i in range(self.n_sim)]
        self.state_ranges = [list(np.arange(i, i + self.n_sub)) for i in start_idx]

        # 7-6. A list of time it took to get the weights equilibrated
        self.equil = [-1 for i in range(self.n_sim)]   # -1 means unequilibrated

        # 7-7. Some variables for counting
        self.n_rejected = 0
        self.n_swap_attempts = 0
        self.n_empty_swappable = 0

        # 7-8. Replica space trajectories. For example, rep_trajs[0] = [0, 2, 3, 0, 1, ...] means
        # that configuration 0 transitioned to replica 2, then 3, 0, 1, in iterations 1, 2, 3, 4, ...,
        # respectively. The first element of rep_traj[i] should always be i.
        self.rep_trajs = [[i] for i in range(self.n_sim)]

        # 7-9. configs shows the current configuration that each replica is sampling.
        # For example, self.configs = [0, 2, 1, 3] means that configurations 0, 2, 1, and 3 are
        # in replicas, 0, 1, 2, 3, respectively. This list will be constantly updated during the simulation.
        self.configs = list(range(self.n_sim))

        # 7-10. The time series of the (processed) whole-range alchemical weights
        # If no weight combination is applied, self.g_vecs will just be a list of None's.
        self.g_vecs = []

        # 7-11. Data analysis
        if self.df_method == 'MBAR':
            self.df_data_type = 'u_nk'
        else:
            self.df_data_type = 'dhdl'

        # 7-12. External module for coordinate modification
        if self.modify_coords is not None:
            sys.path.append(os.getcwd())
            module = importlib.import_module(self.modify_coords)
            self.modify_coords_fn = getattr(module, self.modify_coords)
        else:
            self.modify_coords_fn = None

        # Step 8. Check the executables
        if analysis is False:
            self.check_gmx_executable()

    def check_gmx_executable(self):
        """
        Checks if the GROMACS executable can be used and gets its absolute path and version.
        """
        try:
            result = subprocess.run(['which', self.gmx_executable], capture_output=True, text=True, check=True)
            self.gmx_path = result.stdout.strip()  # this can be exactly the same as self.gmx_executable

            version_output = subprocess.run([self.gmx_path, "-version"], capture_output=True, text=True, check=True)
            for line in version_output.stdout.splitlines():
                if "GROMACS version" in line:
                    self.gmx_version = line.split()[-1]
                    break
        except subprocess.CalledProcessError:
            print(f"{self.gmx_executable} is not available on this system.")
        except Exception as e:
            print(f"An error occurred:\n{e}")

    def print_params(self, params_analysis=False):
        """
        Prints important parameters related to the EXEE simulation.

        Parameters
        ----------
        params_analysis : bool, optional
            If True, additional parameters related to data analysis will be printed. Default is False.
        """
        if isinstance(self.gro, list):
            gro_str = ', '.join(self.gro)
        else:
            gro_str = self.gro

        if isinstance(self.top, list):
            top_str = ', '.join(self.top)
        else:
            top_str = self.top

        print("Important parameters of EXEE")
        print("============================")
        print(f"Python version: {sys.version}")
        print(f"GROMACS executable: {self.gmx_path}")  # we print the full path here
        print(f"GROMACS version: {self.gmx_version}")
        print(f"ensemble_md version: {ensemble_md.__version__}")
        print(f'Simulation inputs: {gro_str}, {top_str}, {self.mdp}')
        print(f"Verbose log file: {self.verbose}")
        print(f"Proposal scheme: {self.proposal}")
        print(f"Whether to perform weight combination: {self.w_combine}")
        print(f"Type of means for weight combination: {self.w_mean_type}")
        print(f"Whether to perform histogram correction: {self.hist_corr}")
        print(f"Histogram cutoff for weight correction: {self.N_cutoff}")
        print(f"Number of replicas: {self.n_sim}")
        print(f"Number of iterations: {self.n_iter}")
        print(f"Length of each replica: {self.dt * self.nst_sim} ps")
        print(f"Frequency for checkpointing: {self.n_ckpt} iterations")
        print(f"Total number of states: {self.n_tot}")
        print(f"Additionally defined swappable states: {self.add_swappables}")
        print(f"Additional grompp arguments: {self.grompp_args}")
        print(f"Additional runtime arguments: {self.runtime_args}")
        # print(f"Number of attempted swaps in one exchange interval: {self.n_ex}")
        if self.mdp_args is not None and len(self.mdp_args.keys()) > 1:
            print("MDP parameters differing across replicas:")
            for i in self.mdp_args.keys():
                print(f"  - {i}: {self.mdp_args[i]}")
        else:
            print(f"MDP parameters differing across replicas: {self.mdp_args}")
        print("Alchemical ranges of each replica in REXEE:")
        for i in range(self.n_sim):
            print(f"  - Replica {i}: States {self.state_ranges[i]}")

        if params_analysis is True:
            print()
            print(f"Whether to build Markov state models and perform relevant analysis: {self.msm}")
            print(f"Whether to perform free energy calculations: {self.free_energy}")
            print(f"The step to used in subsampling the DHDL data in free energy calculations, if any: {self.df_spacing}")  # noqa: E501
            print(f"The chosen free energy estimator for free energy calculations, if any: {self.df_method}")
            print(f"The method for estimating the uncertainty of free energies in free energy calculations, if any: {self.err_method}")  # noqa: E501
            print(f"The number of bootstrap iterations in the boostrapping method, if used: {self.n_bootstrap}")
            print(f"The random seed to use in bootstrapping, if used: {self.seed}")

        if self.reformatted_mdp is True:
            print("Note that the input MDP file has been reformatted by replacing hypens with underscores. The original mdp file has been renamed as *backup.mdp.")  # noqa: E501

    @staticmethod
    def reformat_MDP(mdp_file):
        """
        Reformats the input MDP file so that all hyphens in the parameter names are replaced by underscores.
        If the input MDP file contains hyphens in its parameter names, the class attribue :code:`self.reformatted`
        will be set to :code:`True`. In this case, the new MDP object with reformatted parameter names will be
        written to the original file path of the file, while the original file will be renamed with a
        :code:`_backup` suffix. If the input MDP file is not reformatted, the function sets
        the class attribute :code:`self.reformatted_mdp` to :code:`False`.

        Returns
        -------
        reformatted : bool
            Whether the file was reformatted.
        """
        params = gmx_parser.MDP(mdp_file)

        odict = OrderedDict([(k.replace('-', '_'), v) for k, v in params.items()])
        params_new = gmx_parser.MDP(None, **odict)
        reformatted = None

        if rank == 0:
            if params_new.keys() == params.keys():
                reformatted = False  # no need to reformat the file
            else:
                reformatted = True
                new_name = mdp_file.split('.mdp')[0] + '_backup.mdp'
                os.rename(mdp_file, new_name)
                params_new.write(mdp_file)

        return reformatted

    def initialize_MDP(self, idx):
        """
        Initializes the MDP object for generating MDP files for a replica based on the MDP template.
        This function should be called only for generating MDP files for the FIRST iteration
        and it has nothing to do with whether the weights are fixed or equilibrating.
        It is assumed that the MDP template has all the common parameters of all replicas.

        Parameters
        ----------
        idx : int
            Index of the simulation whose MDP parameters need to be initialized.

        Returns
        -------
        MDP : :obj:`.gmx_parser.MDP` obj
            An updated object of :obj:`.gmx_parser.MDP` that can be used to write MDP files.
        """
        MDP = copy.deepcopy(self.template)
        MDP["nsteps"] = self.nst_sim

        if self.mdp_args is not None:
            for param in self.mdp_args.keys():
                MDP[param] = self.mdp_args[param][idx]

        start_idx = idx * self.s
        for i in self.lambda_types:
            MDP[i] = self.template[i][start_idx:start_idx + self.n_sub]

        if "init_lambda_weights" in self.template:
            MDP["init_lambda_weights"] = self.template["init_lambda_weights"][start_idx:start_idx + self.n_sub]

        return MDP

    def get_ref_dist(self):
        """
        Gets the reference distance(s) to use starting from the second iteration if distance restraint(s) are used.
        Specifically, a reference distance determined here is the initial COM distance between the pull groups
        in the input GRO file. This function initializes the attribute :code:`ref_dist`.
        """
        if hasattr(self, 'set_ref_dist'):
            self.ref_dist = []
            pullx_file = 'sim_0/iteration_0/pullx.xvg'
            for i in range(len(self.set_ref_dist)):
                if self.set_ref_dist[i] is True:
                    # dist = list(extract_dataframe(pullx_file, headers=headers)[f'{i+1}'])[0]
                    dist = np.loadtxt(pullx_file, comments=['#', '@'])[0][1]
                    self.ref_dist.append(dist)

    def update_MDP(self, new_template, sim_idx, iter_idx, states, wl_delta, weights, counts=None):
        """
        Updates the MDP file for a new iteration based on the new MDP template coming from the previous iteration.
        Note that if the weights got equilibrated in the previous iteration, then the weights will be fixed
        at these equilibrated values for all the following iterations.

        Parameters
        ----------
        new_template : str
            The new MDP template file. Typically the MDP file of the previous iteration.
        sim_idx : int
            The index of the simulation whose MDP parameters need to be updated.
        iter_idx : int
            The index of the iteration to be performed later.
        states : list
            A list of last sampled states of all simulaitons in the previous iteration.
        wl_delta : list
            A list of final Wang-Landau incrementors of all simulations.
        weights : list
            A list of lists final weights of all simulations.
        counts : list
            A list of lists final counts of all simulations. If the value is :code:`None`,
            then the MDP parameter :code:`init-histogram-counts` won't be specified in the next iteration.
            Note that not all the GROMACS versions have the MDP parameter :code:`init-histogram-counts` available,
            in which case one should always pass :code:`None`, or set :code:`-maxwarn` in :code:`grompp_args`
            in the input YAML file.

        Return
        ------
        MDP : :obj:`.gmx_parser.MDP` obj
            An updated object of :obj:`.gmx_parser.MDP` that can be used to write MDP files.
        """
        new_template = gmx_parser.MDP(new_template)  # turn into a gmx_parser.MDP object
        MDP = copy.deepcopy(new_template)
        MDP["tinit"] = self.nst_sim * self.dt * iter_idx
        MDP["nsteps"] = self.nst_sim
        MDP["init_lambda_state"] = (states[sim_idx] - sim_idx * self.s)  # 2nd term is for shifting from the global to local index.  # noqa: E501
        MDP["init_wl_delta"] = wl_delta[sim_idx]
        MDP["init_lambda_weights"] = weights[sim_idx]
        if counts is not None:
            MDP["init_histogram_counts"] = counts[sim_idx]

        if self.equil[sim_idx] == -1:   # the weights haven't been equilibrated
            MDP["init_wl_delta"] = wl_delta[sim_idx]
        else:
            MDP["lmc_stats"] = "no"
            MDP["wl_scale"] = ""
            MDP["wl_ratio"] = ""
            MDP["init_wl_delta"] = ""
            MDP["lmc_weights_equil"] = ""
            MDP["weight_equil_wl_delta"] = ""

        # Here we deal with the distance restraint in the pull code, if any.
        if hasattr(self, 'ref_dist'):
            for i in range(len(self.ref_dist)):
                MDP[f'pull_coord{i+1}_start'] = "no"
                MDP[f'pull_coord{i+1}_init'] = self.ref_dist[i]

        return MDP

    def extract_final_dhdl_info(self, dhdl_files):
        """
        For all the replica simulations, finds the last sampled state
        and print the corresponding lambda values from a dhdl file.

        Parameters
        ----------
        dhdl_files : list
            A list of file paths to GROMACS DHDL files of different replicas.

        Returns
        -------
        states : list
            A list of the global indices of the last sampled states of all simulaitons.
        """
        states = []
        print("\nBelow are the final states being visited:")
        for i in range(len(dhdl_files)):
            headers = get_headers(dhdl_files[i])
            state_local = list(extract_dataframe(dhdl_files[i], headers=headers)['Thermodynamic state'])[-1]  # local index of the last state  # noqa: E501
            state_global = state_local + i * self.s  # global index of the last state
            states.append(state_global)  # append the global index
            print(f"  Simulation {i}: Global state {states[i]}")
        print('\n', end='')

        return states

    def extract_final_log_info(self, log_files):
        """
        Extracts the following information for each replica simulation from a given list of GROMACS LOG files:

          - The final Wang-Landau incrementors.
          - The final lists of weights.
          - The final lists of counts.
          - Whether the weights were equilibrated in the simulations.

        Parameters
        ----------
        log_files : list
            A list of file paths to GROMACS LOG files of different replicas.

        Returns
        -------
        wl_delta : list
            A list of final Wang-Landau incrementors of all simulations.
        weights : list
            A list of lists of final weights of all simulations.
        counts : list
            A list of lists of final counts of all simulations.
        """
        wl_delta, weights, counts = [], [], []

        # Find the final Wang-Landau incrementors and weights
        for i in range(len(log_files)):
            if self.verbose:
                print(f'Parsing {log_files[i]} ...')
            result = gmx_parser.parse_log(log_files[i])  # weights, counts, wl_delta, equil_time
            weights.append(result[0][-1])
            counts.append(result[1])
            wl_delta.append(result[2])

            # In Case 3 described in the docstring of parse_log (fixed-weights),
            # result[3] will be 0 but it will never be passed to self.equil[i]
            # because once self.equil[i] is not -1, we stop updating. This way, we can keep
            # the time when the weights get equilibrated all the way.
            if self.equil[i] == -1:
                # At this point self.equil is the equilibration status BEFORE the last iteration
                # while result[3] is the equilibration status AFTER finishing the last iteraiton.
                # For any replicas where weights are still equilibrating (i.e. self.equil[j] == -1)
                # we update its equilibration status.
                self.equil[i] = result[3]

            if self.equilibrated_weights[i] == []:
                if self.equil[i] != -1 and self.equil[i] != 0:
                    # self.equil[i] != -1: uneqilibrated
                    # self.equil[i] != 0: fixed-weight simulation
                    self.equilibrated_weights[i] = result[0][-1]

        return wl_delta, weights, counts

    @staticmethod
    def identify_swappable_pairs(states, state_ranges, neighbor_exchange, add_swappables=None):
        """
        Identify swappable pairs. By definition, a pair of simulation is considered swappable only if
        their last sampled states are in the alchemical ranges of both simulations. This is required
        to ensure that the values of involved ΔH and Δg can always be looked up from the DHDL and LOG files.
        This also automatically guarantee that the simulations to be swapped have overlapping alchemical ranges.

        Parameters
        ----------
        states : list
            A list of the global indices of the last sampled states of all simulations. This list can be
            generated by the :obj:`.extract_final_dhdl_info` method. Notably, the input list should not be
            a list that has been updated/modified by :obj:`get_swapping_pattern`, or the result will be incorrect.
        state_ranges : list of lists
            A list of state indies for all replicas. The input list can be a list updated by
            :obj:`.get_swapping_pattern`, especially in the case where there is a need to re-identify the
            swappable pairs after an attempted swap is accepted.
        neighbor_exchange : bool
            Whether to exchange only between neighboring replicas.
        add_swappables: list
            A list of lists that additionally consider states (in global indices) that can be swapped.
            For example, :code:`add_swappables=[[4, 5], [14, 15]]` means that if a replica samples state 4,
            it can be swapped with another replica that samples state 5 and vice versa. The same logic applies
            to states 14 and 15.

        Returns
        -------
        swappables : list
            A list of tuples representing the simulations that can be swapped.
        """
        n_sim = len(states)
        sim_idx = list(range(n_sim))
        all_pairs = list(combinations(sim_idx, 2))

        # First, we identify pairs of replicas with overlapping ranges
        swappables = [i for i in all_pairs if set(state_ranges[i[0]]).intersection(set(state_ranges[i[1]])) != set()]  # noqa: E501

        # Then, from these pairs, we exclude the ones whose the last sampled states are not present in both alchemical ranges  # noqa: E501
        # In this case, U^i_n, U_^j_m, g^i_n, and g_^j_m are unknown and the acceptance cannot be calculated.
        swappables = [i for i in swappables if states[i[0]] in state_ranges[i[1]] and states[i[1]] in state_ranges[i[0]]]  # noqa: E501

        # Expand the definition of swappable pairs when add_swappables is specified
        if add_swappables is not None:
            all_paired_states = [[states[p[0]], states[p[1]]] for p in all_pairs]
            for i in all_paired_states:
                if i in add_swappables:
                    pair = all_pairs[all_paired_states.index(i)]
                    if pair not in swappables:
                        swappables.append(pair)

        if neighbor_exchange is True:
            print('Note: One neighboring swap will be proposed.')
            swappables = [i for i in swappables if np.abs(i[0] - i[1]) == 1]

        return swappables

    @staticmethod
    def propose_swap(swappables):
        """
        Proposes a swap of coordinates between replicas by drawing samples from the swappable pairs.

        Parameters
        ----------
        swappables : list
            A list of tuples representing the simulations that can be swapped.

        Return
        ------
        swap : tuple or an empty list
            A tuple of simulation indices to be swapped. If there is no swappable pair,
            an empty list will be returned.
        """
        try:
            swap = random.choices(swappables, k=1)[0]
        except IndexError:  # no swappable pairs
            swap = []

        return swap

    def get_swapping_pattern(self, dhdl_files, states):
        """
        Generates a list (:code:`swap_pattern`) that represents how the configurations should be swapped in the
        next iteration. The indices of the output list correspond to the simulation/replica indices, and the
        values represent the configuration indices in the corresponding simulation/replica. For example, if the
        swapping pattern is :code:`[0, 2, 1, 3]`, it means that in the next iteration, replicas 0, 1, 2, 3 should
        sample configurations 0, 2, 1, 3, respectively, where configurations 0, 1, 2, 3 here are defined as whatever
        configurations are in replicas 0, 1, 2, 3 in the CURRENT iteration (not iteration 0), respectively.

        Notably, when this function is called (e.g. once every iteration in an REXEE simulation), the output
        list :code:`swap_pattern` is always initialized as :code:`[0, 1, 2, 3, ...]` and gets updated once every
        attempted swap. This is different from the attribute :code:`configs`, which is only initialized at the
        very beginning of the entire REXEE simulation (iteration 0), though :code:`configs` also gets updated with
        :code:`swap_pattern` once every attempted swap in this function.

        Parameters
        ----------
        dhdl_files : list
            A list of DHDL files. The indicies in the DHDL filenames shouuld be in an ascending order, e.g.
            :code:`[dhdl_0.xvg, dhdl_1.xvg, ..., dhdl_N.xvg]`.
        states : list
            A list of last sampled states (in global indices) of ALL simulaitons. :code:`states[i]=j` means that
            the configuration in replica :code:`i` is at state :code:`j` at the time when the exchange is performed.
            This list can be generated :obj:`.extract_final_dhdl_info`.

        Returns
        -------
        swap_pattern : list
            A list showing the configuration of replicas after swapping.
        swap_list : list
            A list of tuples showing the accepted swaps.
        """
        swap_list = []
        if self.proposal == 'exhaustive':
            n_ex = int(np.floor(self.n_sim / 2))  # This is the maximum, not necessarily the number that will always be reached.  # noqa
            n_ex_exhaustive = 0    # The actual number of swaps atttempted.
        else:
            n_ex = 1  # single swap or neighboring swap

        """
        # multiple swaps
        if self.n_ex == 'N^3':
            n_ex = self.n_tot ** 3
        else:
            n_ex = self.n_ex
        """

        shifts = list(self.s * np.arange(self.n_sim))
        swap_pattern = list(range(self.n_sim))   # Can be regarded as the indices of DHDL files/configurations
        state_ranges = copy.deepcopy(self.state_ranges)
        # states_copy = copy.deepcopy(states)  # only for re-identifying swappable pairs given updated state_ranges --> was needed for the multiple exchange proposal scheme  # noqa: E501
        swappables = ReplicaExchangeEE.identify_swappable_pairs(states, state_ranges, self.proposal == 'neighboring', self.add_swappables)  # noqa: E501

        # Note that if there is only 1 swappable pair, then it will still be the only swappable pair
        # after an attempted swap is accepted. Therefore, there is no need to perform multiple swaps or re-identify
        # the new set of swappable pairs. In this case, we set n_ex back to 1.
        if len(swappables) == 1:
            if n_ex > 1:
                n_ex = 1  # n_ex is set back to 1 since there is only 1 swappable pair.

        print(f"Swappable pairs: {swappables}")
        for i in range(n_ex):
            # Update the list of swappable pairs starting from the 2nd attempted swap for the exhaustive swap method.
            if self.proposal == 'exhaustive' and i >= 1:
                # Note that this should be done regardless of the acceptance/rejection of the previously drawn pairs.
                # Also note that at this point, swap is still the last attempted swap.
                swappables = [i for i in swappables if set(i).intersection(set(swap)) == set()]  # noqa: F821
                print(f'\nRemaining swappable pairs: {swappables}')

            if len(swappables) == 0 and self.proposal == 'exhaustive':
                # This should only happen when the method of exhaustive swaps is used.
                if i == 0:
                    self.n_empty_swappable += 1
                break
            else:
                self.n_swap_attempts += 1
                if self.proposal == 'exhaustive':
                    n_ex_exhaustive += 1

                swap = ReplicaExchangeEE.propose_swap(swappables)
                print(f'\nProposed swap: {swap}')
                if swap == []:
                    self.n_empty_swappable += 1
                    print('No swap is proposed because there is no swappable pair at all.')
                    break  # no need to re-identify swappable pairs and draw new samples
                else:
                    if self.verbose is True and self.proposal != 'exhaustive':
                        print(f'A swap ({i + 1}/{n_ex}) is proposed between the configurations of Simulation {swap[0]} (state {states[swap[0]]}) and Simulation {swap[1]} (state {states[swap[1]]}) ...')  # noqa: E501

                    if self.modify_coords_fn is not None:
                        swap_bool = True  # always accept the move
                    else:
                        # Calculate the acceptance ratio and decide whether to accept the swap.
                        prob_acc = self.calc_prob_acc(swap, dhdl_files, states, shifts)
                        swap_bool = self.accept_or_reject(prob_acc)

                    # Theoretically, in an REXEE simulation, we could either choose to swap configurations (via
                    # swapping GRO files) or replicas (via swapping MDP files). In ensemble_md package, we chose the
                    # former when implementing the REXEE algorithm. Specifically, in the CLI `run_REXEE`,
                    # `swap_pattern` is used to swap the GRO files. Therefore, when an attempted swap is accetped and
                    # `swap_pattern` is updated, we also need to update the variables `shifts`, `weights`,
                    # `dhdl_files`, `state_ranges`, `self.configs` but not anything else. Otherwise, incorrect results
                    # will be produced. To better understand this, one can refer to our unit test for
                    # get_swapping_pattern and calc_prob_acc, set checkpoints and examine why the variables
                    # should/should not be updated.

                    if swap_bool is True:
                        swap_list.append(swap)
                        # The assignments need to be done at the same time in just one line.
                        # states[swap[0]], states[swap[1]] = states[swap[1]], states[swap[0]]
                        # weights[swap[0]], weights[swap[1]] = weights[swap[1]], weights[swap[0]]
                        shifts[swap[0]], shifts[swap[1]] = shifts[swap[1]], shifts[swap[0]]
                        dhdl_files[swap[0]], dhdl_files[swap[1]] = dhdl_files[swap[1]], dhdl_files[swap[0]]
                        swap_pattern[swap[0]], swap_pattern[swap[1]] = swap_pattern[swap[1]], swap_pattern[swap[0]]
                        state_ranges[swap[0]], state_ranges[swap[1]] = state_ranges[swap[1]], state_ranges[swap[0]]
                        self.configs[swap[0]], self.configs[swap[1]] = self.configs[swap[1]], self.configs[swap[0]]

                        """
                        if n_ex > 1 and self.proposal == 'multiple':  # must be multiple swaps
                            # After state_ranges have been updated, we re-identify the swappable pairs.
                            # Notably, states_copy (instead of states) should be used. (They could be different.)
                            swappables = ReplicaExchangeEE.identify_swappable_pairs(states_copy, state_ranges, self.proposal == 'neighboring', self.add_swappables)  # noqa: E501
                            print(f"  New swappable pairs: {swappables}")
                        """
                    else:
                        # In this case, there is no need to update the swappables
                        pass

                print(f'  Current list of configurations: {self.configs}')

        if self.verbose is False:
            print(f'\n{n_ex} swap(s) have been proposed.')
        print(f'\nThe finally adopted swap pattern: {swap_pattern}')
        print(f'The list of configurations sampled in each replica in the next iteration: {self.configs}')

        # Update the replica-space trajectories
        for i in range(self.n_sim):
            self.rep_trajs[i].append(self.configs.index(i))

        return swap_pattern, swap_list

    def calc_prob_acc(self, swap, dhdl_files, states, shifts):
        """
        Calculates the acceptance ratio given the Monte Carlo scheme for swapping the simulations.

        Parameters
        ----------
        swap : tuple
            A tuple of indices corresponding to the simulations to be swapped.
        dhdl_files : list
            A list of DHDL files, e.g. :code:`dhdl_files = ['dhdl_2.xvg', 'dhdl_1.xvg', 'dhdl_0.xvg', 'dhdl_3.xvg']`
            means that configurations 2, 1, 0, and 3 are now in replicas 0, 1, 2, 3. This can happen in multiple swaps
            when a previous swap between configurations 0 and 2 has just been accepted. Otherwise, the list of
            filenames should always be in the ascending order, e.g. :code:`['dhdl_0.xvg', 'dhdl_1.xvg', 'dhdl_2.xvg',
            dhdl_3.xvg]`.
        states : list
            A list of last sampled states (in global indices) in the DHDL files corresponding to configurations 0, 1,
            2, ... (e.g. :code:`dhdl_0.xvg`, :code:`dhdl_1.xvg`, :code:`dhdl_2.xvg`, ...)
            This list can be generated by :obj:`.extract_final_dhdl_info`.
        shifts : list
            A list of state shifts for converting global state indices to the local ones. Specifically, :code:`states`
            substracted by :code:`shifts` should be the local state indices of the last sampled states
            in :code:`dhdl_files[0]`, :code:`dhdl_files[1]`, ... (which importantly, are not necessarily
            :code:`dhdl_0.xvg`, :code:`dhdl_1.xvg`, ...). If multiple swaps are not used, then
            this should always be :code:`list(REXEE.s * np.arange(REXEE.n_sim))`.

        Returns
        -------
        prob_acc : float
            The acceptance ratio.
        """
        if self.verbose is True:
            print("  Proposing a move from (x^i_m, x^j_n) to (x^i_n, x^j_m) ...")
        f0, f1 = dhdl_files[swap[0]], dhdl_files[swap[1]]
        h0, h1 = get_headers(f0), get_headers(f1)
        data_0, data_1 = (
            extract_dataframe(f0, headers=h0).iloc[-1],
            extract_dataframe(f1, headers=h1).iloc[-1],
        )

        # \Delta H to all states at the last time frame
        # Notably, the can be regarded as H for each state since the reference state will have a value of 0 anyway.
        dhdl_0 = data_0[-self.n_sub:]
        dhdl_1 = data_1[-self.n_sub:]

        # Old local index, which will only be used in "metropolis"
        old_state_0 = states[swap[0]] - shifts[swap[0]]
        old_state_1 = states[swap[1]] - shifts[swap[1]]

        # New local index. Note that states are global indices, so we shift them back to the local ones
        new_state_0 = states[swap[1]] - shifts[swap[0]]  # new state index (local index in simulation swap[0])
        new_state_1 = states[swap[0]] - shifts[swap[1]]  # new state index (local index in simulation swap[1])

        dU_0 = (dhdl_0[new_state_0] - dhdl_0[old_state_0]) / self.kT  # U^{i}_{n} - U^{i}_{m}, i.e. \Delta U (kT) to the new state  # noqa: E501
        dU_1 = (dhdl_1[new_state_1] - dhdl_1[old_state_1]) / self.kT  # U^{j}_{m} - U^{j}_{n}, i.e. \Delta U (kT) to the new state  # noqa: E501
        dU = dU_0 + dU_1
        if self.verbose is True:
            print(
                f"  U^i_n - U^i_m = {dU_0:.2f} kT, U^j_m - U^j_n = {dU_1:.2f} kT, Total dU: {dU:.2f} kT"
            )

        prob_acc = min(1, np.exp(-dU))

        return prob_acc

    def accept_or_reject(self, prob_acc):
        """
        Returns a boolean variable indiciating whether the proposed swap should be acceepted given the acceptance rate.

        Parameters
        ----------
        prob_acc : float
            The acceptance rate.

        Returns
        -------
        swap_bool : bool
            A boolean variable indicating whether the swap should be accepted.
        """
        if prob_acc == 0:
            swap_bool = False
            self.n_rejected += 1
            if self.verbose is True:
                print("  Swap rejected! ", end="", flush=True)
        else:
            rand = random.random()
            if self.verbose is True:
                print(
                    f"  Acceptance rate: {prob_acc:.3f} / Random number drawn: {rand:.3f}"
                )
            if rand < prob_acc:
                swap_bool = True
                if self.verbose is True:
                    print("  Swap accepted! ")
            else:
                swap_bool = False
                self.n_rejected += 1
                if self.verbose is True:
                    print("  Swap rejected! ")
        return swap_bool

    def get_averaged_weights(self, log_files):
        """
        For each replica, calculate the averaged weights (and the associated error) from the time series
        of the weights since the previous update of the Wang-Landau incrementor.

        Parameters
        ----------
        log_files : list
            A list of file paths to GROMACS LOG files of different replicas.

        Returned
        --------
        weights_avg : list
            A list of lists of weights averaged since the last update of the Wang-Landau
            incrementor. The length of the list should be the number of replicas.
        weights_err : list
            A list of lists of errors corresponding to the values in :code:`weights_avg`.
        """
        for i in range(self.n_sim):
            weights, _, wl_delta, _ = gmx_parser.parse_log(log_files[i])
            if self.current_wl_delta[i] == wl_delta:
                self.updating_weights[i] += weights  # expand the list
            else:
                self.current_wl_delta[i] = wl_delta
                self.updating_weights[i] = weights

        # shape of self.updating_weights is (n_sim, n_points, n_states), but n_points can be different
        # for different replicas, which will error out np.mean(self.updating_weights, axis=1)
        weights_avg = [np.mean(self.updating_weights[i], axis=0).tolist() for i in range(self.n_sim)]
        weights_err = []
        for i in range(self.n_sim):
            if len(self.updating_weights[i]) == 1:  # this would lead to a RunTime Warning and nan
                weights_err.append([0] * self.n_sub)  # in `weighted_mean``, a simple average will be returned.
            else:
                weights_err.append(np.std(self.updating_weights[i], axis=0, ddof=1).tolist())

        return weights_avg, weights_err

    def weight_correction(self, weights, counts):
        """
        Corrects the lambda weights based on the histogram counts. Namely,
        :math:`g_k' = g_k + ln(N_{k-1}/N_k)`, where :math:`g_k` and :math:`g_k'`
        are the lambda weight after and before the correction, respectively.
        Notably, in any of the following situations, we don't do any correction.

        - Either :math:`N_{k-1}` or :math:`N_k` is 0.
        - Either :math:`N_{k-1}` or :math:`N_k` is smaller than the histogram cutoff.

        Parameters
        ----------
        weights : list
            A list of lists of weights (of ALL simulations) to be corrected.
        counts : list
            A list of lists of counts (of ALL simulations).

        Return
        ------
        weights : list
            An updated list of lists of corected weights.
        """
        skip_correction = False
        for i in range(len(weights)):  # loop over the replicas
            if self.verbose is True:
                print(f"  Counts of rep {i}:\t\t{counts[i]}")
                print(f'  Original weights of rep {i}:\t{[float(f"{k:.3f}") for k in weights[i]]}')

            for j in range(1, len(weights[i])):  # loop over the alchemical states
                if counts[i][j - 1] != 0 and counts[i][j - 1] != 0:
                    if np.min([counts[i][j - 1], counts[i][j]]) > self.N_cutoff:
                        weights[i][j] += np.log(counts[i][j - 1] / counts[i][j])
                    else:
                        skip_correction = True
                        print('Weight correction was deactivated because neither N_{k-1} or N_k is larger than the histogram cutoff.')  # noqa: E501

            if self.verbose is True and skip_correction is False:
                print(f'  Corrected weights of rep {i}:\t{[float(f"{k:.3f}") for k in weights[i]]}\n')

        if self.verbose is False:
            print(' Done')

        return weights

    def histogram_correction(self, hist, print_values=True):
        """
        Adjust the histogram counts. Specifically, the ratio of corrected histogram counts
        for adjancent states is the geometric mean of the ratio of the original histogram counts
        for the same states. Note, however, if the histogram counts are 0 for some states, the
        histogram correction will be skipped and the original histogram counts will be returned.

        Parameters
        ----------
        hist : list
            A list of lists of histogram counts of ALL simulations.
        print_values : bool, optional
            Whether to print the histograms for each replica before and after histogram correction.

        Returns
        -------
        hist_modified : list
            A list of lists of modified histogram counts of ALL simulations.
        """
        # (1) Print the original histogram counts
        if print_values is True:
            print('  Original histogram counts:')
            for i in range(len(hist)):
                print(f'    Rep {i}: {hist[i]}')

        # (2) Calculate adjacent weight differences and g_vec
        N_ratio_vec = []  # N_{k-1}/N_k for the whole range
        with warnings.catch_warnings():  # Suppress the specific warning here
            warnings.simplefilter("ignore", category=RuntimeWarning)
            N_ratio_adjacent = [list(np.array(hist[i][1:]) / np.array(hist[i][:-1])) for i in range(len(hist))]

        for i in range(self.n_tot - 1):
            N_ratio_list = []
            for j in range(len(self.state_ranges)):
                if i in self.state_ranges[j] and i + 1 in self.state_ranges[j]:
                    idx = self.state_ranges[j].index(i)
                    N_ratio_list.append(N_ratio_adjacent[j][idx])
            N_ratio_vec.append(np.prod(N_ratio_list) ** (1 / len(N_ratio_list)))  # geometric mean
        N_ratio_vec.insert(0, hist[0][0])

        # (3) Check if the histogram counts are 0 for some states, if so, the histogram correction will be skipped.
        # Zero histogram counts can happen when the sampling is poor or the WL incrementor just got updated
        contains_nan = any(np.isnan(value) for sublist in N_ratio_adjacent for value in sublist)  # can be caused by 0/0  # noqa: E501
        contains_inf = any(np.isinf(value) for sublist in N_ratio_adjacent for value in sublist)  # can be caused by x/0, where x is a finite number  # noqa: E501
        skip_hist_correction = contains_nan or contains_inf
        if skip_hist_correction:
            print('\n  Histogram correction is skipped because the histogram counts are 0 for some states.')

        # (4) Perform histogram correction if it is not skipped
        if skip_hist_correction is False:
            print('\n  Performing histogram correction ...')
            # When skip_hist_correction is True, previous lines for calculating N_ratio_vec or N_ratio_list will
            # still not error out so it's fine to not add the conditional statement like here, since we will
            # have hist_modified = hist at the end anyway. However, if skip_hist_correction, things like
            # int(np.nan) will lead to an error, so we put an if condition here.
            N_vec = np.array([int(np.ceil(np.prod(N_ratio_vec[:(i + 1)]))) for i in range(len(N_ratio_vec))])

        if skip_hist_correction is False:
            hist_modified = [list(N_vec[self.state_ranges[i]]) for i in range(self.n_sim)]
        else:
            hist_modified = hist  # the original input histogram

        # (5) Print the modified histogram counts
        if print_values is True:
            print('\n  Modified histogram counts:')
            for i in range(len(hist_modified)):
                print(f'    Rep {i}: {hist_modified[i]}')

        return hist_modified

    def combine_weights(self, weights, weights_err=None, print_values=True):
        """
        Combine alchemical weights across multiple replicas. Note that if
        :code:`weights_err` is provided, inverse-variance weighting will be used.
        Care must be taken since inverse-variance weighting can lead to slower
        convergence if the provided errors are not accurate. (See :ref:`doc_w_schemes` for mor details.)

        Parameters
        ----------
        weights : list
            A list of lists alchemical weights of ALL simulations.
        weights_err : list, optional
            A list of lists of errors corresponding to the values in :code:`weights`.
        print_values : bool, optional
            Whether to print the weights for each replica before and
            after weight combination for each replica.

        Returns
        -------
        weights_modified : list
            A list of modified Wang-Landau weights of ALL simulations.
        g_vec : np.ndarray
            An array of alchemical weights of the whole range of states.
        """
        # (1) Print the original weights
        if print_values is True:
            w = np.round(weights, decimals=3).tolist()  # just for printing
            print('  Original weights:')
            for i in range(len(w)):
                print(f'    Rep {i}: {w[i]}')

        # (2) Calculate adjacent weight differences and g_vec
        dg_vec = []  # alchemical weight differences for the whole range
        dg_adjacent = [list(np.diff(weights[i])) for i in range(len(weights))]

        if weights_err is not None:
            dg_adjacent_err = [[np.sqrt(weights_err[i][j] ** 2 + weights_err[i][j + 1] ** 2) for j in range(len(weights_err[i]) - 1)] for i in range(len(weights_err))]  # noqa: E501

        for i in range(self.n_tot - 1):
            dg_list, dg_err_list = [], []
            for j in range(len(self.state_ranges)):
                if i in self.state_ranges[j] and i + 1 in self.state_ranges[j]:
                    idx = self.state_ranges[j].index(i)
                    dg_list.append(dg_adjacent[j][idx])
                    if weights_err is not None:
                        dg_err_list.append(dg_adjacent_err[j][idx])
            if weights_err is None:
                dg_vec.append(np.mean(dg_list))
            else:
                dg_vec.append(utils.weighted_mean(dg_list, dg_err_list)[0])

        dg_vec.insert(0, 0)
        g_vec = np.array([sum(dg_vec[:(i + 1)]) for i in range(len(dg_vec))])

        # (3) Determine the vector of alchemical weights for each replica
        weights_modified = np.zeros_like(weights)
        for i in range(self.n_sim):
            if self.equil[i] == -1:  # unequilibrated
                weights_modified[i] = list(g_vec[i * self.s: i * self.s + self.n_sub] - g_vec[i * self.s: i * self.s + self.n_sub][0])  # noqa: E501
            else:
                weights_modified[i] = self.equilibrated_weights[i]

        # (4) Print the modified weights
        if print_values is True:
            w = np.round(weights_modified, decimals=3).tolist()  # just for printing
            print('\n  Modified weights:')
            for i in range(len(w)):
                print(f'    Rep {i}: {w[i]}')

        if self.verbose is False:
            print(' DONE')
            print(f'The alchemical weights of all states: \n  {list(np.round(g_vec, decimals=3))}')
        else:
            print(f'\n  The alchemical weights of all states: \n  {list(np.round(g_vec, decimals=3))}')

        return weights_modified, g_vec

    def _run_grompp(self, n, swap_pattern):
        """
        Prepares TPR files for the simulation ensemble using the GROMACS :code:`grompp` command.

        Parameters
        ----------
        n : int
            The iteration index (starting from 0).
        swap_pattern : list
            A list generated by :obj:`.get_swapping_pattern`. It represents how the replicas should be swapped.
        """
        args_list = []
        for i in range(self.n_sim):
            arguments = [self.gmx_executable, 'grompp']

            # Input files
            mdp = f"sim_{i}/iteration_{n}/{self.mdp.split('/')[-1]}"
            if n == 0:
                if isinstance(self.gro, list):
                    gro = f"{self.gro[i]}"
                else:
                    gro = f"{self.gro}"
            else:
                gro = f"sim_{swap_pattern[i]}/iteration_{n-1}/confout.gro"  # This effectively swap out GRO files

            if isinstance(self.top, list):
                top = f"{self.top[i]}"
            else:
                top = f"{self.top}"

            # Add input file arguments
            arguments.extend(["-f", mdp, "-c", gro, "-p", top])

            # Add output file arguments
            arguments.extend([
                "-o", f"sim_{i}/iteration_{n}/sys_EE.tpr",
                "-po", f"sim_{i}/iteration_{n}/mdout.mdp"
            ])

            # Add additional arguments if any
            if self.grompp_args is not None:
                # Turn the dictionary into a list with the keys alternating with values
                add_args = [elem for pair in self.grompp_args.items() for elem in pair]
                arguments.extend(add_args)

            args_list.append(arguments)

        # Run the GROMACS grompp commands in parallel
        returncode = None  # Initialize as None for all ranks (necessary for the case when -np > n_sim, which is rare)
        if rank == 0:
            print('Generating TPR files ...')
        if rank < self.n_sim:
            returncode, stdout, stderr = utils.run_gmx_cmd(args_list[rank])
            if returncode != 0:
                print(f'Error on rank {rank} (return code: {returncode}):\n{stderr}')

        # gather return codes at rank 0
        code_list = comm.gather(returncode, root=0)

        if rank == 0:
            # Filter out None values which represent ranks that did not execute the command
            code_list = [code for code in code_list if code is not None]
            if code_list != [0] * self.n_sim:
                MPI.COMM_WORLD.Abort(1)   # Doesn't matter what non-zero returncode we put here as the code from GROMACS will be printed before this point anyway.  # noqa: E501

    def _run_mdrun(self, n):
        """
        Executes GROMACS mdrun commands in parallel.

        Parameters
        ----------
        n : int
            The iteration index (starting from 0).
        """
        # We will change the working directory so the mdrun command should be the same for all replicas.
        arguments = [self.gmx_executable, 'mdrun']

        # Add input file arguments
        arguments.extend(['-s', 'sys_EE.tpr'])

        if self.runtime_args is not None:
            # Turn the dictionary into a list with the keys alternating with values
            add_args = [elem for pair in self.runtime_args.items() for elem in pair]
            arguments.extend(add_args)

        # Run the GROMACS mdrun commands in parallel
        returncode = None  # Initialize as None for all ranks (necessary for the case when -np > n_sim, which is rare)
        if rank == 0:
            print('Running EXE simulations ...')
        if rank < self.n_sim:
            os.chdir(f'sim_{rank}/iteration_{n}')
            returncode, stdout, stderr = utils.run_gmx_cmd(arguments)
            if returncode != 0:
                print(f'Error on rank {rank} (return code: {returncode}):\n{stderr}')
            if self.rm_cpt is True:
                # if the simulation went wrong, there would be no checkpoint file
                try:
                    os.remove('state.cpt')
                except Exception:
                    print('\n--------------------------------------------------------------------------\n')
                    MPI.COMM_WORLD.Abort(1)
            os.chdir('../../')

        # gather return codes at rank 0
        code_list = comm.gather(returncode, root=0)

        if rank == 0:
            # Filter out None values which represent ranks that did not execute the command
            code_list = [code for code in code_list if code is not None]
            if code_list != [0] * self.n_sim:
                MPI.COMM_WORLD.Abort(1)   # Doesn't matter what non-zero returncode we put here as the code from GROMACS will be printed before this point anyway.  # noqa: E501

    def run_REXEE(self, n, swap_pattern=None):
        """
        Perform one iteration in the REXEE simulation, which includes generating the
        TPR files using the GROMACS grompp :code:`command` and running the expanded ensemble simulations
        in parallel using GROMACS :code:`mdrun` command. The GROMACS commands are launched by as subprocesses.
        The function assumes that the GROMACS executable is available.

        Parameters
        ----------
        n : int
            The iteration index (starting from 0).
        swap_pattern : list
            A list generated by :obj:`.get_swapping_pattern`. It represents how the replicas should be swapped.
            This parameter is not needed only if :code:`n` is 0.
        """
        if rank == 0:
            iter_str = f'\nIteration {n}: {self.dt * self.nst_sim * n: .1f} - {self.dt * self.nst_sim * (n + 1): .1f} ps'  # noqa: E501
            print(iter_str + '\n' + '=' * (len(iter_str) - 1))

        # 1st synchronizing point for all MPI processes: To make sure ranks other than 0 will not start executing
        # _run_grompp earlier and mess up the order of printing.
        comm.barrier()

        # Generating all required TPR files simultaneously, then run all simulations simultaneously.
        # No synchronizing point is needed between _run_grompp and _run_mdrun, since once rank i finishes _run_grompp,
        # it should run _run_mdrun in the same working directory, so there won't be any I/O error.
        self._run_grompp(n, swap_pattern)
        self._run_mdrun(n)

        # 2nd synchronizaing point for all MPI processes: To make sure no rank will start getting to the next
        # iteration earlier than the others. For example, if rank 0 finishes the mdrun command earlier, we don't
        # want it to start parsing the dhdl file (in the if condition of if rank == 0) of simulation 3 being run by
        # rank 3 that has not been generated, which will lead to an I/O error.
        comm.barrier()



####################################################################
#                                                                  #
#    ensemble_md,                                                  #
#    a python package for running GROMACS simulation ensembles     #
#                                                                  #
#    Written by Wei-Tse Hsu <wehs7661@colorado.edu>                #
#    Copyright (c) 2022 University of Colorado Boulder             #
#                                                                  #
####################################################################
"""
The :code:`gmx_parser` module provides functions for parsing GROMACS files.
"""
import os
import re
import six
import logging
import warnings
from collections import OrderedDict as odict

from ensemble_md.utils import utils
from ensemble_md.utils.exceptions import ParseError


def parse_log(log_file):
    """
    This function parses a log file generated by expanded ensemble and provides
    important information, especially for running new iterations in REXEE.
    Typically, there are three types of log files from an expanded ensemble simulation:

    - **Case 1**: The weights are still updating in the simulation and have never been equilibrated.

        - The output :code:`equil_time` should always be -1.

    - **Case 2**: The weights were equilibrated during the simulation.

        - The output :code:`equil_time` should be the time (in ps) it took to get the weights equilibrated.
        - The final weights (:code:`weights`) will just be the equilibrated weights.

    - **Case 3**: The weights were fixed in the simulation.

        - The output :code:`equil_time` should always be 0.
        - The final weights (which never change during the simulation) and the final counts will
          still be returned.

    Parameters
    ----------
    log_file : str
        The log file to be parsed.

    Returns
    -------
    weights : list
        In all cases, :code:`weights` should be a list of lists (of weights).

        - In Case 1, a list of list of weights as a function of time since the last update of the Wang-Landau
          incrementor will be returned.
        - In Case 2, a list of list of weights as a function of time since the last update of the Wang-Landau
          incrementor up to equilibration will be returned.
        - In Case 3, the returned list will only have one list inside, which is the list of the final weights.

        That is, for all cases, :code:`weights[-1]` will be the final weights, which are useful in REXEE.
    counts : list
        The final histogram counts.
    wl_delta : float
        The final Wang-Landau incementor. In Cases 2 and 3, :code:`None` will be returned.
    equil_time : int or float
        - In Case 1, -1 will be returned, which means that the weights have not been equilibrated.
        - In Case 2, the time in ps that it took to equilibrate the weights will be returned.
        - In Case 3, 0 will be returned, which means that the weights were fixed during the simulation.
    """
    f = open(log_file, "r")
    lines = f.readlines()
    f.close()

    case = None  # itialized as None and should end up as 1 or 2 or 3.

    # First parse the MD parameters to tell the type of the simulation.
    for l in lines:  # noqa: E741
        if "n-lambdas" in l:
            N_states = int(l.split("=")[1])
        if "tinit" in l:
            tinit = float(l.split("=")[1])
        if "weight-equil-wl-delta" in l:
            # For Case 1 and Case 2
            cutoff = float(l.split("=")[1])
        if "lmc-stats" in l:
            if l.split("=")[1].split()[0] in ["no", "No"]:  # Case 3
                case = '3'
                equil_time = 0
                wl_delta = None
            else:
                pass  # Either Case 1 or Case 2
        if "dt  " in l:
            dt = float(l.split("=")[1])

    # For all cases, we need to find weights and counts
    weights = []
    if case == '3':  # could be either '3' or None
        # We only need the info at the end of the simulation, so it's faster to search from the bottom of the file.
        lines.reverse()
        n = -1
        counts = []
        for l in lines:  # noqa: E741
            n += 1
            if "Count   G(in kT)" in l:
                w = []  # the list of weights at this time frame
                # The first occurrence would be the final weights. (We've reversed the lines!)
                for i in range(1, N_states + 1):
                    if "<<" in lines[n - i]:
                        w.append(float(lines[n - i].split()[-3]))
                        counts.append(int(lines[n - i].split()[-4]))
                    else:
                        w.append(float(lines[n - i].split()[-2]))
                        counts.append(int(lines[n - i].split()[-3]))
                weights.append(w)
                break
    else:  # Case 1 and Case 2
        # Here we search from the top, since we need weights as a function of time anyway.
        n = -1
        find_equil, append_equil = False, False
        wl_delta_list = [None]  # We use None so that the change in wl-delta will always get deteced at the beginning
        for l in lines:  # noqa: E741
            n += 1
            if "Count   G(in kT)" in l:  # this line is lines[n]
                # We should first check if wl_delta is changed, set weights=[] if needed before we append new weights
                if "Wang-Landau incrementor is" in lines[n - 1]:  # Case 1 so far: wl_delta is still updating
                    case = '1'
                    equil_time = -1
                    current_wl_delta = float(lines[n - 1].split(":")[1])
                    if wl_delta_list[-1] != current_wl_delta and current_wl_delta > cutoff:  # when wl-delta changes
                        # Note that if the time step the equilibration is reached happens to be a multiple of nstlog,
                        # a wl-delta right below the cutoff will be printed. In that case, we don't want to reset
                        # `weights` so we get the time series since the last time when we still have wl_delta > cutoff
                        weights = []  # we want only time series since the latest update of wl_delta
                    wl_delta_list.append(float(lines[n - 1].split(":")[1]))

                w, counts = [], []  # the list of weights at this time frame
                for i in range(1, N_states + 1):
                    # counts will be constantly updated by new values
                    if "<<" in lines[n + i]:
                        w.append(float(lines[n + i].split()[-3]))
                        counts.append(int(lines[n + i].split()[-4]))
                    else:
                        w.append(float(lines[n + i].split()[-2]))
                        counts.append(int(lines[n + i].split()[-3]))

                if find_equil is False or append_equil is False:
                    weights.append(w)
                    if find_equil is True:
                        append_equil = True

            if "Weights have equilibrated" in l:
                case = '2'  # we don't break the loop even if equil_time is found, as we need the final counts.
                find_equil = True  # After this, we will append weights one last time, which are equilibrated weights.
                equil_step = int(l.split(":")[0].split("Step")[1])
                equil_time = equil_step * dt + tinit  # ps
                if wl_delta is not None and wl_delta < cutoff:
                    # Should only happen when equil_time % nstlog == 0, where the weights should have been appended
                    # Note that we additionally have wl_delta is not None. Since wl_delta could be None if the
                    # weights get equilibrated right after the simulation stop (before nstlog).
                    append_equil = True

            wl_delta = wl_delta_list[-1]

        if case == '2':
            wl_delta = None

    return weights, counts, wl_delta, equil_time


class FileUtils(object):
    """Mixin class to provide additional file-related capabilities.
    Modified from `utilities.py in GromacsWrapper <https://github.com/Becksteinlab/GromacsWrapper>`_.
    Copyright (c) 2009 Oliver Beckstein <orbeckst@gmail.com>
    """

    #: Default extension for files read/written by this class.
    default_extension = None

    def _init_filename(self, filename=None, ext=None):
        """Initialize the current filename :attr:`FileUtils.real_filename` of the object.

        Bit of a hack.

        - The first invocation must have ``filename != None``; this will set a
          default filename with suffix :attr:`FileUtils.default_extension`
          unless another one was supplied.

        - Subsequent invocations either change the filename accordingly or
          ensure that the default filename is set with the proper suffix.

        """

        extension = ext or self.default_extension
        filename = self.filename(
            filename, ext=extension, use_my_ext=True, set_default=True
        )
        #: Current full path of the object for reading and writing I/O.
        self.real_filename = os.path.realpath(filename)

    def filename(self, filename=None, ext=None, set_default=False, use_my_ext=False):
        """Supply a file name for the class object.

        Typical uses::

           fn = filename()             ---> <default_filename>
           fn = filename('name.ext')   ---> 'name'
           fn = filename(ext='pickle') ---> <default_filename>'.pickle'
           fn = filename('name.inp','pdf') --> 'name.pdf'
           fn = filename('foo.pdf',ext='png',use_my_ext=True) --> 'foo.pdf'

        The returned filename is stripped of the extension
        (``use_my_ext=False``) and if provided, another extension is
        appended. Chooses a default if no filename is given.

        Raises a ``ValueError`` exception if no default file name is known.

        If ``set_default=True`` then the default filename is also set.

        ``use_my_ext=True`` lets the suffix of a provided filename take
        priority over a default ``ext`` tension.
        """
        if filename is None:
            if not hasattr(self, "_filename"):
                self._filename = None  # add attribute to class
            if self._filename:
                filename = self._filename
            else:
                raise ValueError(
                    "A file name is required because no default file name was defined."
                )
            my_ext = None
        else:
            filename, my_ext = os.path.splitext(filename)
            if set_default:  # replaces existing default file name
                self._filename = filename
        if my_ext and use_my_ext:
            ext = my_ext
        if ext is not None:
            if ext.startswith(os.extsep):
                ext = ext[1:]  # strip a dot to avoid annoying mistakes
            if ext != "":
                filename = filename + os.extsep + ext
        return filename


class MDP(odict, FileUtils):
    """Class that represents a Gromacs mdp run input file.
    Modified from `GromacsWrapper <https://github.com/Becksteinlab/GromacsWrapper>`_.
    Copyright (c) 2009-2011 Oliver Beckstein <orbeckst@gmail.com>
    The MDP instance is an ordered dictionary.

      - *Parameter names* are keys in the dictionary.
      - *Comments* are sequentially numbered with keys Comment0001,
        Comment0002, ...
      - *Empty lines* are similarly preserved as Blank0001, ....

    When writing, the dictionary is dumped in the recorded order to a
    file. Inserting keys at a specific position is not possible.

    Currently, comments after a parameter on the same line are
    discarded. Leading and trailing spaces are always stripped.
    """

    default_extension = "mdp"
    logger = logging.getLogger("gromacs.formats.MDP")

    COMMENT = re.compile("""\s*;\s*(?P<value>.*)""")  # eat initial ws  # noqa: W605
    # see regex in cbook.edit_mdp()
    PARAMETER = re.compile(
        """
                            \s*(?P<parameter>[^=]+?)\s*=\s*  # parameter (ws-stripped), before '='  # noqa: W605
                            (?P<value>[^;]*)                # value (stop before comment=;)  # noqa: W605
                            (?P<comment>\s*;.*)?            # optional comment  # noqa: W605
                            """,
        re.VERBOSE,
    )

    def __init__(self, filename=None, autoconvert=True, **kwargs):
        """Initialize mdp structure.

        :Arguments:
          *filename*
              read from mdp file
          *autoconvert* : boolean
              ``True`` converts numerical values to python numerical types;
              ``False`` keeps everything as strings [``True``]
          *kwargs*
              Populate the MDP with key=value pairs. (NO SANITY CHECKS; and also
              does not work for keys that are not legal python variable names such
              as anything that includes a minus '-' sign or starts with a number).
        """
        super(MDP, self).__init__(
            **kwargs
        )  # can use kwargs to set dict! (but no sanity checks!)

        self.autoconvert = autoconvert

        if filename is not None:
            self._init_filename(filename)
            self.read(filename)

    def __eq__(self, other):
        """
        __eq__ inherited from FileUtils needs to be overridden if new attributes (autoconvert in
        this case) are assigned to the instance of the subclass (MDP in our case).
        See `this post by LGTM <https://lgtm.com/rules/9990086/>`_ for more details.
        """
        if not isinstance(other, MDP):
            return False
        return FileUtils.__eq__(self, other) and self.autoconvert == other.autoconvert

    def _transform(self, value):
        if self.autoconvert:
            return utils._autoconvert(value)
        else:
            return value.rstrip()

    def read(self, filename=None):
        """Read and parse mdp file *filename*."""
        self._init_filename(filename)

        def BLANK(i):
            return "B{0:04d}".format(i)

        def COMMENT(i):
            return "C{0:04d}".format(i)

        data = odict()
        iblank = icomment = 0
        with open(self.real_filename) as mdp:
            for line in mdp:
                line = line.strip()
                if len(line) == 0:
                    iblank += 1
                    data[BLANK(iblank)] = ""
                    continue
                m = self.COMMENT.match(line)
                if m:
                    icomment += 1
                    data[COMMENT(icomment)] = m.group("value")
                    continue
                # parameter
                m = self.PARAMETER.match(line)
                if m:
                    # check for comments after parameter?? -- currently discarded
                    parameter = m.group("parameter")
                    value = self._transform(m.group("value"))
                    data[parameter] = value
                else:
                    errmsg = "{filename!r}: unknown line in mdp file, {line!r}".format(
                        **vars()
                    )
                    self.logger.error(errmsg)
                    raise ParseError(errmsg)

        super(MDP, self).update(data)

    def write(self, filename=None, skipempty=False):
        """Write mdp file to *filename*.

        Parameters
        ----------
        filename : str
            Output mdp file; default is the filename the mdp was read from. If the filename
            is not supplied, the function will overwrite the file that the mdp was read from.
        skipempty : bool
            ``True`` removes any parameter lines from output that contain empty values [``False``]
        """
        # The line 'if skipempty and (v == "" or v is None):' below could possibly incur FutureWarning
        warnings.simplefilter(action='ignore', category=FutureWarning)

        with open(self.filename(filename, ext="mdp"), "w") as mdp:
            for k, v in self.items():
                if k[0] == "B":  # blank line
                    mdp.write("\n")
                elif k[0] == "C":  # comment
                    mdp.write("; {v!s}\n".format(**vars()))
                else:  # parameter = value
                    if skipempty and (v == "" or v is None):
                        continue
                    if isinstance(v, six.string_types) or not hasattr(v, "__iter__"):
                        mdp.write("{k!s} = {v!s}\n".format(**vars()))
                    else:
                        mdp.write("{} = {}\n".format(k, " ".join(map(str, v))))


def compare_MDPs(mdp_list, print_diff=False):
    """
    Given a list of MDP files, identify the parameters for which not all MDP
    files have the same values. Note that this function is not aware of the default
    values of GROMACS parameters. (Currently, this function is not used in the
    workflow adopted in :code:`run_REXEE.py` but it might be useful in some places,
    so we decided to keep it.)

    Parameters
    ----------
    mdp_list : list
        A list of MDP files.
    print_diff : bool
        If :code:`True`, print to screen the parameters that are different among the MDP files
        and the values of the parameters in the MDP files in a more readable format.

    Returns
    -------
    diff_params : dict
        A dictionary of parameters that are different among the MDP files.
        The keys are the parameter names and the values is a list of values of the
        parameters in the MDP files.
    """
    diff_params = {}
    for i in range(len(mdp_list)):
        mdps = [MDP(mdp_list[i]) for i in range(len(mdp_list))]
        params_dicts = [odict([(k.replace('-', '_'), v) if type(v) is not str else (k.replace('-', '_'), v.replace('-', '_')) for k, v in p.items()]) for p in mdps]  # noqa: E501

        # First figure out the union set of the parameters and exclude blanks and comments
        all_params = set([key for d in params_dicts for key in d.keys()])
        all_params = [p for p in all_params if not p.startswith(('B', 'C'))]

        for p in all_params:
            if p in diff_params:
                pass  # already in the dictionary, no need to compare again
            else:
                if not all(p in d for d in params_dicts):
                    # the parameter is not in all MDP files
                    diff_params[p] = [d[p] if p in d else None for d in params_dicts]
                else:
                    # the parameter is in all MDP files (Note that "set([1, 1, 1]={1}.)")
                    if isinstance(params_dicts[0][p], list):
                        # the parameter is a list, which is unhashable
                        if len(set([tuple(d[p]) for d in params_dicts])) > 1:
                            diff_params[p] = [d[p] for d in params_dicts]
                    else:
                        if len(set([d[p] for d in params_dicts])) > 1:
                            diff_params[p] = [d[p] for d in params_dicts]

    if print_diff:
        print("The following parameters are different among the MDP files:")
        for k, v in diff_params.items():
            print(k)
            for i in range(len(mdp_list)):
                print(f'  - {mdp_list[i]}: {v[i]}')
            print()

    return diff_params



####################################################################
#                                                                  #
#    ensemble_md,                                                  #
#    a python package for running GROMACS simulation ensembles     #
#                                                                  #
#    Written by Wei-Tse Hsu <wehs7661@colorado.edu>                #
#    Copyright (c) 2022 University of Colorado Boulder             #
#                                                                  #
####################################################################
"""
The :obj:`.analyze_traj` module provides methods for analyzing trajectories in REXEE.
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator

from alchemlyb.parsing.gmx import _get_headers as get_headers
from alchemlyb.parsing.gmx import _extract_dataframe as extract_dataframe
from ensemble_md.utils import utils


def extract_state_traj(dhdl):
    """
    Extracts the state-space trajectory from a GROMACS DHDL file.
    Note that the state indices here are local indices.

    Parameters
    ----------
    dhdl : str
        The filename of the GROMACS DHDL file to be parsed.

    Returns
    -------
    traj : list
        A list that represents that state-space trajectory
    t : list
        A list that represents the time series of the trajectory
    """
    traj = list(extract_dataframe(dhdl, headers=get_headers(dhdl))['Thermodynamic state'])
    t = list(np.loadtxt(dhdl, comments=['#', '@'])[:, 0])

    return traj, t


def stitch_time_series(files, rep_trajs, shifts=None, dhdl=True, col_idx=-1, save_npy=True):
    """
    Stitches the state-space/CV-space trajectories for each starting configuration from DHDL files
    or PLUMED output files generated at different iterations.

    Parameters
    ----------
    files : list
        A list of lists of file names of GROMACS DHDL files or general GROMACS XVG files or PLUMED ouptput files.
        Specifically, :code:`files[i]` should be a list containing the files of interest from all iterations in
        replica :code:`i`. The files should be sorted naturally.
    rep_trajs : list
        A list of lists that represents the replica space trajectories for each starting configuration. For example,
        :code:`rep_trajs[0] = [0, 2, 3, 0, 1, ...]` means that starting configuration 0 transitioned to replica 2, then
        3, 0, 1, in iterations 1, 2, 3, 4, ..., respectively.
    shifts : list
        A list of values for shifting the state indices for each replica. The length of the list
        should be equal to the number of replicas. This is only needed when :code:`dhdl=True`.
    dhdl : bool
        Whether the input files are GROMACS dhdl files, in which case trajectories of global alchemical indices
        will be generated. If :code:`dhdl=False`, the input files must be readable by `numpy.loadtxt` assuming that
        the start of a comment is indicated by either the :code:`#` or :code:`@` characters.
        Such files include any GROMACS XVG files or PLUMED output files (output by plumed driver, for instance).
        In this case, trajectories of the configurational collective variable of interest are generated.
        The default is :code:`True`.
    col_idx : int
        The index of the column to be extracted from the input files. This is only needed when :code:`dhdl=False`,
        By default, we extract the last column.
    save_npy : bool
        Whether to save the output trajectories as an NPY file.

    Returns
    -------
    trajs : list
        A list that contains lists of state-space/CV-space trajectory (in global indices) for each starting
        configuration. For example, :code:`trajs[i]` is the state-space/CV-space trajectory of starting
        configuration :code:`i`.
    """
    n_configs = len(files)  # number of starting configurations
    n_iter = len(files[0])  # number of iterations per replica

    # First figure out which dhdl/plumed output files each starting configuration corresponds to
    # files_sorted[i] contains the dhdl/plumed output files for starting configuration i sorted
    # based on iteration indices
    files_sorted = [[] for i in range(n_configs)]
    for i in range(n_configs):
        for j in range(n_iter):
            files_sorted[i].append(files[rep_trajs[i][j]][j])

    # Then, stitch the trajectories for each starting configuration
    trajs = [[] for i in range(n_configs)]  # for each starting configuration
    for i in range(n_configs):
        for j in range(n_iter):
            if j == 0:
                if dhdl:
                    traj, t = extract_state_traj(files_sorted[i][j])
                else:
                    traj = np.loadtxt(files_sorted[i][j], comments=['#', '@'])[:, col_idx]
                    t = np.loadtxt(files_sorted[i][j], comments=['#', '@'])[:, 0]  # only used if save_xvg is True
            else:
                # Starting from the 2nd iteration, we get rid of the first time frame the first
                # frame of iteration n+1 the is the same as the last frame of iteration n
                if dhdl:
                    traj, t = extract_state_traj(files_sorted[i][j])
                    traj, t = traj[1:], t[1:]
                else:
                    traj = np.loadtxt(files_sorted[i][j], comments=['#', '@'])[:, col_idx][1:]

            if dhdl:  # Trajectories of global alchemical indices will be generated.
                shift_idx = rep_trajs[i][j]
                traj = list(np.array(traj) + shifts[shift_idx])
            trajs[i].extend(traj)

    if save_npy is True:
        if dhdl:
            np.save('state_trajs.npy', trajs)
        else:
            np.save('cv_trajs.npy', trajs)

    return trajs


def convert_npy2xvg(trajs, dt, subsampling=1):
    """
    Convert a :code:`state_trajs.npy` or :code:`cv_trajs.npy` file to :math:`N_{\text{rep}}` XVG files
    that have two columns: time (ps) and state index.

    Parameters
    ----------
    trajs : ndarray
        The state-space or CV-space trajectories read from :code:`state_trajs.npy` or :code:`cv_trajs.npy`.
    dt : float
        The time interval (in ps) between consecutive frames of the trajectories.
    subsampling : int
        The stride for subsampling the time series. The default is 1.
    """
    n_configs = len(trajs)
    for i in range(n_configs):
        traj = trajs[i]
        t = np.arange(len(traj)) * dt
        headers = ['This file was created by ensemble_md']
        if 'int' in str(traj.dtype):
            headers.extend(['Time (ps) v.s. State index'])
            np.savetxt(f'traj_{i}.xvg', np.transpose([t[::subsampling], traj[::subsampling]]), header='\n'.join(headers), fmt=['%-8.1f', '%4.0f'])  # noqa: E501
        else:
            headers.extend(['Time (ps) v.s. CV'])
            np.savetxt(f'traj_{i}.xvg', np.transpose([t[::subsampling], traj[::subsampling]]), header='\n'.join(headers), fmt=['%-8.1f', '%8.6f'])  # noqa: E501


def stitch_time_series_for_sim(files, shifts=None, dhdl=True, col_idx=-1, save=True):
    """
    Stitches the state-space/CV-space time series in the same replica/simulation folder.
    That is, the output time series is contributed by multiple different trajectories (initiated by
    different starting configurations) to a certain alchemical range.

    Parameters
    ----------
    files : list
        A list of lists of file names of GROMACS DHDL files or general GROMACS XVG files
        or PLUMED output files. Specifically, :code:`files[i]` should be a list containing
        the files of interest from all iterations in replica :code:`i`. The files should be sorted naturally.
    shifts : list
        A list of values for shifting the state indices for each replica. The length of the list
        should be equal to the number of replicas. This is only needed when :code:`dhdl=True`.
    dhdl : bool
        Whether the input files are GROMACS dhdl files, in which case trajectories of global alchemical indices
        will be generated. If :code:`dhdl=False`, the input files must be readable by `numpy.loadtxt` assuming that
        the start of a comment is indicated by either the :code:`#` or :code:`@` characters.
        Such files include any GROMACS XVG files or PLUMED output files (output by plumed driver, for instance).
        In this case, trajectories of the configurational collective variable of interest are generated.
        The default is :code:`True`.
    col_idx : int
        The index of the column to be extracted from the input files. This is only needed when :code:`dhdl=False`,
        By default, we extract the last column.
    save : bool
        Whether to save the output trajectories as an NPY file.

    Returns
    -------
    trajs : list
        A list that contains lists of state-space/CV-space trajectory (in global indices) for each replica.
        For example, :code:`trajs[i]` is the state-space/CV-space trajectory of replica :code:`i`.
    """
    n_sim = len(files)      # number of replicas
    n_iter = len(files[0])  # number of iterations per replica
    trajs = [[] for i in range(n_sim)]
    for i in range(n_sim):
        for j in range(n_iter):
            if dhdl:
                traj, _ = extract_state_traj(files[i][j])
            else:
                traj = np.loadtxt(files[i][j], comments=['#', '@'])[:, col_idx]

            if dhdl:
                traj = list(np.array(traj) + shifts[i])

            if j != 0:
                traj = traj[:-1]  # remove the last frame, which is the same as the first of the next time series.
            trajs[i].extend(traj)

    # Save the trajectories as an NPY file if desired
    if save is True:
        np.save('state_trajs_for_sim.npy', trajs)

    return trajs


def stitch_trajs(gmx_executable, files, rep_trajs):
    """
    Demuxes GROMACS trajectories from different replicas into individual continuous trajectories.

    Parameters
    ----------
    gmx_executable : str
        The path to the GROMACS executable.
    files : list
        A list of lists of file names of GROMACS XTC files. Specifically, :code:`files[i]` should be a list containing
        the files of interest from all iterations in replica :code:`i`. The files should be sorted naturally.
    rep_trajs : list
        A list of lists that represents the replica space trajectories for each starting configuration. For example,
        :code:`rep_trajs[0] = [0, 2, 3, 0, 1, ...]` means that starting configuration 0 transitioned to replica 2, then
        3, 0, 1, in iterations 1, 2, 3, 4, ..., respectively.
    """
    n_sim = len(files)      # number of replicas
    n_iter = len(files[0])  # number of iterations per replica

    # First figure out which xtc files each starting configuration corresponds to
    # files_sorted[i] contains the xtc files for starting configuration i sorted
    # based on iteration indices
    files_sorted = [[] for i in range(n_sim)]
    for i in range(n_sim):
        for j in range(n_iter):
            files_sorted[i].append(files[rep_trajs[i][j]][j])

    # Then, stitch the trajectories for each starting configuration
    for i in range(n_sim):
        print(f'Recovering the continuous trajectory {i} by concatenating the XTC files ...')
        arguments = [gmx_executable, 'trjcat', '-f']
        arguments.extend(files_sorted[i])
        arguments.extend(['-o', f'traj_{i}.xtc'])
        returncode, stdout, stderr = utils.run_gmx_cmd(arguments)
        if returncode != 0:
            print(f'Error with return code: {returncode}):\n{stderr}')


def traj2transmtx(traj, N, normalize=True):
    """
    Computes the transition matrix given a trajectory. For example, if a state-space
    trajectory from a EXE or HREX simulation given, a state transition matrix is returned.
    If a trajectory showing transitions between replicas in a REXEE simulation is given,
    a replica transition matrix is returned.

    Parameters
    ---------
    traj : list
        A list of state indices showing the trajectory in the state space.
    N : int
        The size (N) of the expcted transition matrix (N by N).
    normalize : bool
        Whether to normalize the matrix so that each row sum to 1. If False, then
        the entries will be the counts of transitions.

    Returns
    -------
    transmtx : np.ndarray
        The transition matrix computed from the trajectory
    """
    transmtx = np.zeros([N, N])
    for i in range(1, len(traj)):
        transmtx[traj[i - 1], traj[i]] += 1   # counts of transitions
    if normalize is True:
        transmtx /= np.sum(transmtx, axis=1)[:, None]   # normalize the transition matrix
        transmtx[np.isnan(transmtx)] = 0   # for non-sampled state, there could be nan due to 0/0

    return transmtx


def plot_rep_trajs(trajs, fig_name, dt=None, stride=None):
    """
    Plots the time series of replicas visited by each trajectory in a single plot.

    Parameters
    ----------
    trajs : list
        A list of arrays that represent the all replica space trajectories.
    fig_name : str
        The file name of the png file to be saved (with the extension).
    dt : float or None, optional
        One trajectory timestep in ps. If None, it assumes there are no timeframes but MC steps.
    stride : int, optional
        The stride for plotting the time series. The default is 100 if the length of
        any trajectory has more than one million frames. Otherwise, it will be 1. Typically
        plotting more than 10 million frames can take a lot of memory.
    """
    n_sim = len(trajs)
    cmap = plt.cm.ocean  # other good options are CMRmap, gnuplot, terrain, turbo, brg, etc.
    colors = [cmap(i) for i in np.arange(n_sim) / n_sim]

    if dt is None:
        x = np.arange(len(trajs[0]))
    else:
        x = np.arange(len(trajs[0])) * dt
        if max(x) >= 10000:
            x = x / 1000   # convert to ns
            units = 'ns'
        else:
            units = 'ps'

    if stride is None:
        if len(trajs[0]) > 1000000:
            stride = 100
        else:
            stride = 1

    fig = plt.figure()
    ax = fig.add_subplot(111)
    for i in range(n_sim):
        if len(trajs[0]) >= 100:  # don't show the markers
            plt.plot(x[::stride], trajs[i][::stride], color=colors[i], label=f'Trajectory {i}')
        else:
            plt.plot(x[::stride], trajs[i][::stride], color=colors[i], label=f'Trajectory {i}', marker='o')

    if dt is None:
        plt.xlabel('MC moves')
    else:
        plt.xlabel(f'Time ({units})')

    plt.ylabel('Replica')
    plt.grid()
    plt.legend()
    ax.yaxis.set_major_locator(MaxNLocator(integer=True))
    plt.savefig(f'{fig_name}', dpi=600)


def plot_state_trajs(trajs, state_ranges, fig_name, dt=None, stride=1, title_prefix='Trajectory'):
    """
    Plots the time series of state index.

    Parameters
    ----------
    trajs : list
        A list of state index time series either from different continuous trajectories or from different
        alchemical ranges (i.e. from different simulation folders).
    state_ranges : list
        A list of lists of state indices. (Like the attribute :code:`state_ranges` in :code:`EnsemblEXE`.)
    fig_name : str
        The file name of the png file to be saved (with the extension).
    dt : float or None, optional
        The time interval between consecutive frames of the trajectories. If None, it is assumed
        that the trajectories are in terms of Monte Carlo (MC) moves instead of timeframes, and
        the x-axis label is set to 'MC moves'. Default is None.
    stride : int
        The stride for plotting the time series. The default is 10 if the length of
        any trajectory has more than 100,000 frames. Otherwise, it will be 1. Typically
        plotting more than 10 million frames can take a lot of memory.
    title_prefix : str
        The prefix shared by the titles of the subplots. For example, if :code:`title_prefix` is
        set to "Trajectory", then the titles of the subplots will be "Trajectory 0", "Trajectory 1", ..., etc.
    """
    n_sim = len(trajs)
    cmap = plt.cm.ocean  # other good options are CMRmap, gnuplot, terrain, turbo, brg, etc.
    colors = [cmap(i) for i in np.arange(n_sim) / n_sim]

    if dt is None:
        x = np.arange(len(trajs[0]))
    else:
        x = np.arange(len(trajs[0])) * dt
        if max(x) >= 10000:
            x = x / 1000   # convert to ns
            units = 'ns'
        else:
            units = 'ps'

    if stride is None:
        if len(trajs[0]) > 100000:
            stride = 10
        else:
            stride = 1

    # x_range = [-5, len(trajs[0]) - 1 + 5]
    x_range = [np.min(x), np.max(x)]
    y_range = [-0.2, np.max(trajs) + 0.2]
    n_configs = len(trajs)
    n_rows, n_cols = utils._get_subplot_dimension(n_configs)
    _, ax = plt.subplots(nrows=n_rows, ncols=n_cols, figsize=(2.5 * n_cols, 2.5 * n_rows))
    for i in range(n_configs):
        plt.subplot(n_rows, n_cols, i + 1)

        # First color different regions showing alchemical ranges
        for j in range(n_configs):
            bounds = [list(state_ranges[j])[0], list(state_ranges[j])[-1]]
            if j == 0:
                bounds[0] -= 0.5
            if j == n_configs - 1:
                bounds[1] += 0.5
            plt.fill_between(x_range, y1=bounds[1], y2=bounds[0], color=colors[j], alpha=0.1)

        if len(trajs[0]) > 100000:
            linewidth = 0.01
        else:
            linewidth = 1  # this is the default

        # Finally, plot the trajectories
        linewidth = 1  # this is the default
        plt.plot(x[::stride], trajs[i][::stride], color=colors[i], linewidth=linewidth)
        if dt is None:
            plt.xlabel('MC moves')
        else:
            plt.xlabel(f'Time ({units})')

        plt.ylabel('State')
        plt.title(f'{title_prefix} {i}', fontweight='bold')
        if len(trajs[0]) >= 10000:
            plt.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))
        plt.xlim(x_range)
        plt.ylim(y_range)
        plt.grid()

    # Remove redundant subplots
    n_rm = n_cols * n_rows - n_configs
    for i in range(n_rm):
        ax.flat[-1 * (i + 1)].set_visible(False)

    plt.tight_layout()
    plt.savefig(f'{fig_name}', dpi=600)


def plot_state_hist(trajs, state_ranges, fig_name, stack=True, figsize=None, prefix='Trajectory', subplots=False, save_hist=True):  # noqa: E501
    """
    Plots state index histograms.

    Parameters
    ----------
    trajs : list
        A list of state index time series either from different continuous trajectories or from different
        alchemical ranges (i.e. from different simulation folders).
    state_ranges : list
        A list of lists of state indices. (Like the attribute :code:`state_ranges` in :obj:`.ReplicaExchangeEE`.)
    fig_name : str
        The file name of the png file to be saved (with the extension).
    stack : bool
        Whether to stack the histograms. Only meaningful when :code:`subplots` is :code:`False`.
    figsize : tuple
        A tuple specifying the length and width of the output figure. The
        default is :code:`(6.4, 4.8)` for cases having less than 30 states and :code:`(10, 4.8)` otherwise.
    prefix : str
        The prefix shared by the titles of the subplots, or the labels shown in the same plot.
        For example, if :code:`prefix` is set to "Trajectory", then the titles/labels of the
        will be "Trajectory 0", "Trajectory 1", ..., etc.
    subplots : bool
        Whether to plot the histogram in multiple subplots, with the title of
        each based on the value of :code:`prefix`.
    save_hist : bool
        Whether to save the histogram data.

    Returns
    -------
    hist_data : list
        The histogram data of the each state index time series.
    """
    n_configs = len(trajs)
    n_states = max(max(state_ranges)) + 1
    cmap = plt.cm.ocean  # other good options are CMRmap, gnuplot, terrain, turbo, brg, etc.
    colors = [cmap(i) for i in np.arange(n_configs) / n_configs]

    hist_data = []
    lower_bound, upper_bound = -0.5, n_states - 0.5
    for traj in trajs:
        hist, bins = np.histogram(traj, bins=np.arange(lower_bound, upper_bound + 1, 1))
        hist_data.append(hist)
    if save_hist is True:
        np.save('hist_data.npy', hist_data)

    # Use the same bins for all histograms
    bins = bins[:-1]  # Remove the last bin edge because there are n+1 bin edges for n bins

    # Start plotting
    if figsize is None:
        if max(trajs[-1]) > 30:
            figsize = (10, 4.8)
        else:
            figsize = (6.4, 4.8)  # default

    fig = plt.figure(figsize=figsize)

    if subplots is False:
        # Initialize the list of bottom (only matters for stack = True)
        bottom = [0] * n_states

        ax = fig.add_subplot(111)
        y_max = 0
        for i in range(n_configs):
            max_count = np.max(bottom + hist_data[i])
            if max_count > y_max:
                y_max = max_count
            plt.bar(
                range(n_states),
                hist_data[i],
                align='center',
                width=1,
                color=colors[i],
                edgecolor='black',
                label=f'{prefix} {i}',
                alpha=0.5,
                bottom=bottom
            )

            if stack is True:
                bottom = [b + c for b, c in zip(bottom, hist_data[i])]

        plt.xticks(range(n_states))

        # Here we color the different regions to show alchemical ranges
        y_max *= 1.05
        for i in range(n_configs):
            bounds = [list(state_ranges[i])[0], list(state_ranges[i])[-1]]
            if i == 0:
                bounds[0] -= 0.5
            if i == n_configs - 1:
                bounds[1] += 0.5
            plt.fill_betweenx([0, y_max], x1=bounds[1] + 0.5, x2=bounds[0] - 0.5, color=colors[i], alpha=0.1, zorder=0)
        plt.xlim([lower_bound, upper_bound])
        plt.ylim([0, y_max])
        plt.xlabel('State index')
        plt.ylabel('Count')
        plt.grid()
        plt.legend()
        plt.tight_layout()
        plt.savefig(f'{fig_name}', dpi=600)
    else:
        n_rows, n_cols = utils._get_subplot_dimension(n_configs)
        _, ax = plt.subplots(nrows=n_rows, ncols=n_cols, figsize=(4 * n_cols, 3 * n_rows))
        for i in range(n_configs):
            plt.subplot(n_rows, n_cols, i + 1)
            plt.bar(
                state_ranges[i],
                hist_data[i][state_ranges[i]],
                align='center',
                width=1,
                edgecolor='black',
                alpha=0.5
            )

            plt.xticks(state_ranges[i], fontsize=8)
            plt.xlim([state_ranges[i][0] - 0.5, state_ranges[i][-1] + 0.5])
            plt.xlabel('State index')
            plt.ylabel('Count')
            plt.title(f'{prefix} {i}')
            plt.grid()
        plt.tight_layout()
        plt.savefig(f'{fig_name}', dpi=600)

    return hist_data


def calculate_hist_rmse(hist_data, state_ranges):
    """
    Calculates the RMSE of accumulated histogram counts of the state index. The reference
    is determined by assuming all alchemical states have equal chances to be visited, i.e.
    the alchemical weights are perfect.

    Parameters
    ----------
    hist_data : list
        The histogram data of the state index for each trajectory.
    state_ranges : list
        A list of lists of state indices. (Like the attribute :code:`state_ranges` in :obj:`.ReplicaExchangeEE`.)

    Returns
    -------
    rmse : float
        The RMSE value of accumulated histogram counts of the state index.
    """
    N = np.max(state_ranges) + 1  # the number of states
    n_accessible = np.histogram(state_ranges, bins=np.arange(-0.5, N + 0.5))[0]
    n_samples = np.sum(hist_data)  # Should be equal to (n_iter * nst_sim / nstdhdl + 1) * n_sim
    n_states_sum = np.sum(n_accessible)  # n_sub * n_sim
    hist_ref = n_samples * (n_accessible / n_states_sum)  # may not be all integers but should be fine
    hist_acc = np.sum(hist_data, axis=0)
    rmse = np.sqrt(np.sum((hist_acc - hist_ref) ** 2) / len(hist_ref))

    return rmse


def plot_transit_time(trajs, N, fig_prefix=None, dt=None, folder='.'):
    """
    Caclulcates and plots the average transit times for each trajectory, including the time
    it takes from states 0 to k, from k to 0 and from 0 to k back to 0 (i.e. round-trip time).
    If there are more than 100 round-trips, 3 histograms corresponding to t_0k, t_k0 and t_roundtrip
    will be generated.

    Parameters
    ----------
    trajs : list
        A list of arrays that represent the state space trajectories of all continuous trajectories.
    N : int
        The total number of states in the whole alchemical range.
    fig_prefix : str
        A prefix to use for all generated figures.
    dt : float or None, optional
        One trajectory timestep in ps. If None, it assumes there are no timeframes but MC steps.
    folder : str, optional
        The directory for saving the figures

    Returns
    -------
    t_0k_list : list
        A list of transit time from states 0 to k for each trajectory.
    t_k0_list : list
        A list of transit time from states k to 0 for each trajectory.
    t_roundtrip_list : list
        A list of round-trip times for each trajectory.
    units : str
        The units of the times.
    """
    if dt is None:
        x = np.arange(len(trajs[0]))
        units = 'step'
    else:
        x = np.arange(len(trajs[0])) * dt
        if max(x) >= 10000:
            x = x / 1000   # convert to ns
            units = 'ns'
        else:
            units = 'ps'

    # The lists below are for storing data corresponding to different trajectories.
    t_0k_list, t_k0_list, t_roundtrip_list = [], [], []
    t_0k_avg, t_k0_avg, t_roundtrip_avg = [], [], []

    sci = False  # whether to use scientific notation in the y-axis in the plot
    t_max = 0  # the maximum time across trajectories --> just for decideing the units
    for i in range(len(trajs)):
        traj = trajs[i]
        last_visited = None   # last visited end
        k = N - 1
        t_0, t_k = [], []   # time frames visting states 0 and k (k is the other end)

        # time spent from statkes 0 to k, k to 0 and the round-trip time (from 0 to k to 0)
        t_0k, t_k0, t_roundtrip = [], [], []

        end_0_found, end_k_found = None, None
        for t in range(len(traj)):
            if traj[t] == 0:
                end_0_found = True
                if last_visited != 0:
                    t_0.append(t)
                    if last_visited == k:
                        t_k0.append(t - t_k[-1])
                last_visited = 0
            if traj[t] == k:
                end_k_found = True
                if last_visited != k:
                    t_k.append(t)
                    if last_visited == 0:
                        t_0k.append(t - t_0[-1])
                last_visited = k

        # Here we figure out the round-trip time from t_0k and t_k0.
        if len(t_0k) != len(t_k0):   # then it must be len(t_0k) = len(t_k0) + 1 or len(t_k0) = len(t_0k) + 1, so we drop the last element of the larger list  # noqa: E501
            if len(t_0k) > len(t_k0):
                t_0k.pop()
            else:
                t_k0.pop()
        t_roundtrip = list(np.array(t_0k) + np.array(t_k0))

        if end_0_found is True and end_k_found is True:
            if dt is not None:
                units = 'ps'
                t_0k = list(np.array(t_0k) * dt)  # units: ps
                t_k0 = list(np.array(t_k0) * dt)  # units: ps
                t_roundtrip = list(np.array(t_roundtrip) * dt)  # units: ps
                if len(t_0k) + len(t_k0) + len(t_roundtrip) > 0:  # i.e. not all are empty
                    if np.max([t_0k, t_k0, t_roundtrip]) > t_max:
                        t_max = np.max([t_0k, t_k0, t_roundtrip])

                    if t_max >= 10000:
                        units = 'ns'
                        t_0k = list(np.array(t_0k) / 1000)   # units: ns
                        t_k0 = list(np.array(t_k0) / 1000)   # units: ns
                        t_roundtrip = list(np.array(t_roundtrip) / 1000)   # units: ns

            t_0k_list.append(t_0k)
            t_0k_avg.append(np.mean(t_0k))

            t_k0_list.append(t_k0)
            t_k0_avg.append(np.mean(t_k0))

            t_roundtrip_list.append(t_roundtrip)
            t_roundtrip_avg.append(np.mean(t_roundtrip))

            if len(t_0k) + len(t_k0) + len(t_roundtrip) > 0:  # i.e. not all are empty
                if sci is False and np.max([t_0k, t_k0, t_roundtrip]) >= 10000:
                    sci = True
        else:
            t_0k_list.append([])
            t_k0_list.append([])
            t_roundtrip_list.append([])

    # Now we plot! (If there are no events, the figures will just be blank)
    meta_list = [t_0k_list, t_k0_list, t_roundtrip_list]
    y_labels = [
        f'Average transit time from states 0 to k ({units})',
        f'Average transit time from states k to 0 ({units})',
        f'Average round-trip time ({units})',
    ]
    fig_names = ['t_0k.png', 't_k0.png', 't_roundtrip.png']
    for t in range(len(meta_list)):
        t_list = meta_list[t]
        if all(not x for x in t_list):
            # If the nested list is empty, no plots will be generated.
            pass
        else:
            len_list = [len(i) for i in t_list]
            if np.max(len_list) <= 10:
                marker = 'o'
            else:
                marker = ''

            plt.figure()
            for i in range(len(t_list)):    # t_list[i] is the list for trajectory i
                plt.plot(np.arange(len(t_list[i])) + 1, t_list[i], label=f'Trajectory {i}', marker=marker)

            if max(max((t_list))) >= 10000:
                plt.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
            plt.xlabel('Event index')
            plt.ylabel(f'{y_labels[t]}')
            plt.grid()
            plt.legend()
            if fig_prefix is None:
                plt.savefig(f'{folder}/{fig_names[t]}')
            else:
                plt.savefig(f'{folder}/{fig_prefix}_{fig_names[t]}', dpi=600)

            lens = [len(t_list[i]) for i in range(len(t_list))]
            if np.min(lens) >= 100:  # plot a histogram
                counts, bins = np.histogram(t_list[i])

                plt.figure()
                for i in range(len(t_list)):
                    plt.hist(t_list[i], bins=int(len(t_list[i]) / 20), label=f'Trajectory {i}')
                    if max(counts) >= 10000:
                        plt.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
                plt.xlabel(f'{y_labels[t]}')
                plt.ylabel('Event count')
                plt.grid()
                plt.legend()
                if fig_prefix is None:
                    plt.savefig(f'{folder}/hist_{fig_names[t]}', dpi=600)
                else:
                    plt.savefig(f'{folder}/{fig_prefix}_hist_{fig_names[t]}', dpi=600)

    return t_0k_list, t_k0_list, t_roundtrip_list, units


def plot_g_vecs(g_vecs, refs=None, refs_err=None, plot_rmse=True):
    """
    Plots the alchemical weight for each alchemical intermediate state as a function of
    the iteration index. Note that the alchemical weight of the first state (which is always 0)
    is skipped. If the reference values are given, they will be plotted in the figure and
    an RMSE will be calculated.

    Parameters
    ----------
    g_vecs : np.array
        The alchemical weights of all states as a function of iteration index. The shape should
        be (n_iterations, n_states). Such an array can be directly read from :code:`g_vecs.npy`
        saved by :code:`run_REXEE`.
    refs : np.array
        The reference values of the alchemical weights.
    refs_err : list or np.array
        The errors of the reference values.
    plot_rmse : bool
        Whether to plot RMSE as a function of the iteration index.
    """
    # n_iter, n_state = g_vecs.shape[0], g_vecs.shape[1]
    g_vecs = np.transpose(g_vecs)
    n_sim = len(g_vecs)
    cmap = plt.cm.ocean  # other good options are CMRmap, gnuplot, terrain, turbo, brg, etc.
    colors = [cmap(i) for i in np.arange(n_sim) / n_sim]
    plt.figure()
    for i in range(1, len(g_vecs)):
        if len(g_vecs[0]) < 100:
            plt.plot(range(len(g_vecs[i])), g_vecs[i], label=f'State {i}', c=colors[i], linewidth=0.8, marker='o', markersize=2)  # noqa: E501
        else:  # plot without markers
            plt.plot(range(len(g_vecs[i])), g_vecs[i], label=f'State {i}', c=colors[i], linewidth=0.8)
    plt.xlabel('Iteration index')
    plt.ylabel('Alchemical weight (kT)')
    ax = plt.gca()
    x_range = ax.get_xlim()
    plt.xlim([0, x_range[1]])
    plt.grid()
    plt.legend(loc='center left', bbox_to_anchor=(1, 0.2))

    if refs is not None:
        for i in range(1, len(refs)):
            plt.axhline(y=refs[i], c='black', linestyle='--', linewidth=0.5)
            if refs_err is not None:
                ax = plt.gca()
                x_range = ax.get_xlim()
                plt.fill_between(x_range, y1=refs[i] - refs_err[i], y2=refs[i] + refs_err[i], color='lightgreen')

        # Calculate the RMSE as a function of the iteration index
        RMSE_list = [np.sqrt(np.mean((g_vecs[:, i] - refs) ** 2)) for i in range(len(g_vecs[0]))]
        plt.text(0.02, 0.95, f'Final RMSE: {RMSE_list[-1]:.3f} kT', transform=ax.transAxes)
        print(f'Final RMSE: {RMSE_list[-1]: .3f} kT')

    plt.tight_layout()
    plt.savefig('g_vecs.png', dpi=600)

    if refs is not None and plot_rmse is True:
        plt.figure()
        plt.plot(range(len(g_vecs[i])), RMSE_list)
        plt.xlabel('Iteration index')
        plt.ylabel('RMSE in the alchemical weights (kT)')
        plt.grid()
        plt.savefig('g_vecs_rmse.png', dpi=600)


def get_swaps(REXEE_log='run_REXEE_log.txt'):
    """
    For each replica, identifies the states where exchanges were proposed and accepted.
    (Todo: We should be able to only use :code:`rep_trajs.npy` and :code:`state_trajs.npy`
    instead of parsing the REXEE log file to reach the same goal.)

    Parameters
    ----------
    REXEE_log : str
        The output log file of the REXEE simulation.

    Returns
    -------
    proposed_swaps : list
        A list of dictionaries showing where the swaps were proposed in
        each replica. Each dictionary (corresponding to one replica) have
        keys being the global state indices and values being the number of
        proposed swaps that occurred in the state indicated by the key.
    accepted_swaps : list
        A list of dictionaries showing where the swaps were accepted in
        each replica. Each dictionary (corresponding to one replica) have
        keys being the global state indices and values being the number of
        accepted swaps that occurred in the state indicated by the key.
    """
    f = open(REXEE_log, 'r')
    lines = f.readlines()
    f.close()

    state_list = []
    for line in lines:
        if 'Number of replicas: ' in line:
            n_sim = int(line.split('Number of replicas: ')[-1])
        if '- Replica' in line:
            state_list.append(eval(line.split('States ')[-1]))

        if 'Iteration' in line:
            break

    # Note that proposed_swaps and accepted_swaps are initialized in the same way
    proposed_swaps = [{i: 0 for i in state_list[j]} for j in range(n_sim)]  # Key: global state index; Value: The number of accepted swaps  # noqa: E501
    accepted_swaps = [{i: 0 for i in state_list[j]} for j in range(n_sim)]  # Key: global state index; Value: The number of accepted swaps  # noqa: E501
    state_trajs = [[] for i in range(n_sim)]  # the state-space trajectory for each REPLICA (not trajectory)
    for line in lines:
        if 'Simulation' in line and 'Global state' in line:
            rep = int(line.split(':')[0].split()[-1])
            state = int(line.split(',')[0].split()[-1])
            state_trajs[rep].append(state)

        if 'Proposed swap' in line:
            swap = eval(line.split(': ')[-1])
            proposed_swaps[swap[0]][state_trajs[swap[0]][-1]] += 1  # states_trajs[swap[0]][-1] is the last state sampled by swap[0]  # noqa: E501
            proposed_swaps[swap[1]][state_trajs[swap[1]][-1]] += 1  # states_trajs[swap[1]][-1] is the last state sampled by swap[1]  # noqa: E501

        if 'Swap accepted!' in line:
            accepted_swaps[swap[0]][state_trajs[swap[0]][-1]] += 1  # states_trajs[swap[0]][-1] is the last state sampled by swap[0]  # noqa: E501
            accepted_swaps[swap[1]][state_trajs[swap[1]][-1]] += 1  # states_trajs[swap[1]][-1] is the last state sampled by swap[1]  # noqa: E501

    return proposed_swaps, accepted_swaps


def plot_swaps(swaps, swap_type='', stack=True, figsize=None):
    """
    Plots the histogram of the proposed swaps or accepted swaps for each replica.

    Parameters
    ----------
    swaps : list
        A list of dictionaries showing showing the number of swaps for each
        state for each replica. This list could be either of the outputs from :obj:`.get_swaps`.
    swap_type : str
        The value should be either :code:`'accepted'` or :code:`'proposed'`. This value
        will only influence the name of y-axis and the output file name.
    stack : bool
        Whether to stack the histograms.
    figsize : tuple
        A tuple specifying the length and width of the output figure. The
        default is :code:`(6.4, 4.8)` for cases having less than 30 states and :code:`(10, 4.8)` otherwise.
    """
    n_sim = len(swaps)
    n_states = max(max(d.keys()) for d in swaps) + 1
    lower_bound, upper_bound = -0.5, n_states - 0.5
    state_ranges = [list(swaps[i].keys()) for i in range(n_sim)]
    cmap = plt.cm.ocean
    colors = [cmap(i) for i in np.arange(n_sim) / n_sim]

    # A new list of dictionaries, each of which consider all state indies
    full_data = [{state: d.get(state, 0) for state in range(n_states)} for d in swaps]  # d.get(state, 0) returns 0 if the state is unavilable  # noqa: E501

    # counts of swaps for all states
    counts_list = [[d[state] for state in range(n_states)] for d in full_data]

    # Initialize the list of bottom
    bottom = [0] * n_states

    if figsize is None:
        if n_states > 30:
            figsize = (10, 4.8)
        else:
            figsize = (6.4, 4.8)  # default

    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)

    for i in range(n_sim):
        plt.bar(
            range(n_states),
            counts_list[i],
            align='center',
            width=1,
            color=colors[i],
            edgecolor='black',
            label=f'Replica {i}',
            alpha=0.5,
            bottom=bottom
        )

        if stack is True:
            bottom = [b + c for b, c in zip(bottom, counts_list[i])]

    plt.xticks(range(n_states))

    # Here we color the different regions to show alchemical ranges
    y_min, y_max = ax.get_ylim()
    for i in range(n_sim):
        bounds = [list(state_ranges[i])[0], list(state_ranges[i])[-1]]
        if i == 0:
            bounds[0] -= 0.5
        if i == n_sim - 1:
            bounds[1] += 0.5
        plt.fill_betweenx([y_min, y_max], x1=bounds[1] + 0.5, x2=bounds[0] - 0.5, color=colors[i], alpha=0.1, zorder=0)
    plt.xlim([lower_bound, upper_bound])
    # plt.ylim([y_min, y_max])
    plt.xlabel('State')
    plt.ylabel(f'Number of {swap_type} swaps')
    plt.grid()
    plt.legend()
    plt.tight_layout()
    if swap_type == '':
        plt.savefig('swaps.png', dpi=600)
    else:
        plt.savefig(f'{swap_type}_swaps.png', dpi=600)


def get_g_evolution(log_files, N_states, avg_frac=0, avg_from_last_update=False):
    """
    For weight-updating simulations, gets the time series of the alchemical
    weights of all states. Note that this funciton is only suitable for analyzing
    either a single expanded ensemble simulation or a replica in a REXEE simulation
    (given all the log files for the replica).

    Parameters
    ----------
    log_files : list
        The list of log file names.
    N_states : int
        The total number of states in the whole alchemical range.
    avg_frac : float
        The fraction of the last part of the simulation to be averaged. The
        default is 0, which means no averaging. Note that this parameter is
        ignored if :code:`avg_from_last_update` is :code:`True`.
    avg_from_last_update : bool
        Whether to average from the last update of wl-delta. If False, the
        averaging will be from the beginning of the simulation.

    Returns
    -------
    g_vecs_all : list
        The alchemical weights of all states as a function of time.
        It should be a list of lists.
    g_vecs_avg : list
        The alchemical weights of all states averaged over the last part of
        the simulation. If :code:`avg_frac` is 0, :code:`None` will be returned.
    g_vecs_err : list
        The errors of the alchemical weights of all states averaged over the
        last part of the simulation. If :code:`avg_frac` is 0 and :code:`avg_from_last_update`
        is :code:`False`, :code:`None` will be returned.
    """
    g_vecs_all = []
    idx_updates = []  # the indices of the data points corresponding to the updates of wl-delta
    for log_file in log_files:
        f = open(log_file, "r")
        lines = f.readlines()
        f.close()

        n = -1
        find_equil = False
        for line in lines:
            n += 1
            if "Count   G(in kT)" in line:  # this line is lines[n]
                w = []  # the list of weights at this time frame
                for i in range(1, N_states + 1):
                    if "<<" in lines[n + i]:
                        w.append(float(lines[n + i].split()[-3]))
                    else:
                        w.append(float(lines[n + i].split()[-2]))

                if find_equil is False:
                    g_vecs_all.append(w)

            if 'weights are now' in line:
                idx_updates.append(len(g_vecs_all) - 1)

            if "Weights have equilibrated" in line:
                find_equil = True
                # Usually, the line two lines above "Weights have been equilibrated" is the line
                # "Step xxx: weights are now: xxx", but there could be exceptions, in which case
                # we just do not append anything since the last fixed weights should have been alreayd appended.
                # The exception happens when the change of the WL incrmentor and happened at the time when
                # the log file is written, in which case one WL incrementor below than the cutoff will be printed,
                # leading to different formats of the log file where "weights are now" is not in lines[n-2].
                if "weights are now:" in lines[n-2]:
                    w = [float(i) for i in lines[n - 2].split(':')[-1].split()]
                    g_vecs_all.append(w)
                break

    if avg_from_last_update is True:
        # If the weights are equilibrated, then the last occurrence of "weights are now"
        # is right before the equilibration message, in which case we want to average
        # from the second last occurrence of "weights are now".
        if find_equil is True:
            idx_updates = idx_updates[:-1]

        idx_last_update = idx_updates[-1]
        g_vecs_avg = np.mean(g_vecs_all[idx_last_update + 1:], axis=0)
        g_vecs_err = np.std(g_vecs_all[idx_last_update + 1:], axis=0, ddof=1)
    else:
        if avg_frac != 0:
            n_avg = int(avg_frac * len(g_vecs_all))
            g_vecs_avg = np.mean(g_vecs_all[-n_avg:], axis=0)
            g_vecs_err = np.std(g_vecs_all[-n_avg:], axis=0, ddof=1)
        else:
            g_vecs_avg = None
            g_vecs_err = None

    return g_vecs_all, g_vecs_avg, g_vecs_err


def get_dg_evolution(log_files, start_state, end_state):
    """
    For weight-updating simulations, gets the time series of the weight
    difference (:math:`Δg = g_2-g_1`) between the specified states.

    Parameters
    ----------
    log_files : list
        The list of log file names.
    start_state : int
        The index of the state (starting from 0) whose weight is :math:`g_1`.
    end_state : int
        The index of the state (starting from 0) whose weight is :math:`g_2`.

    Returns
    -------
    dg : list
        A list of :math:`Δg` values.
    """
    N_states = end_state - start_state + 1  # number of states for the range of insterest
    g_vecs = get_g_evolution(log_files, N_states)
    dg = [g_vecs[i][end_state] - g_vecs[i][start_state] for i in range(len(g_vecs))]

    return dg


def plot_dg_evolution(log_files, start_state, end_state, start_idx=0, end_idx=-1, dt_log=2):
    """
    For weight-updating simulations, plots the time series of the weight
    difference (:math:`Δg = g_2-g_1`) between the specified states.

    Parameters
    ----------
    log_files : list
        The list of log file names.
    start_state : int
        The index of the state (starting from 0) whose weight is :math:`g_1`.
    end_state : int
        The index of the state (starting from 0) whose weight is :math:`g_2`.
    start_idx : int
        The index of the first frame to be plotted.
    end_idx : int
        The index of the last frame to be plotted.
    dt_log : float
        The time interval between two consecutive frames in the log file. The
        default is 2 ps.
    """
    dg = get_dg_evolution(log_files, start_state, end_state)

    # Now we plot
    dg = dg[start_idx:end_idx]
    t = np.arange(len(dg)) * dt_log
    plt.figure()
    if max(t) >= 10000:
        t = t / 1000
        units = 'ns'
    else:
        units = 'ps'
    plt.plot(t, dg)
    plt.xlabel(f'Time ({units})')
    plt.ylabel(r'$\Delta g$')
    plt.grid()
    plt.savefig('dg_evolution.png', dpi=600)

    return dg


def get_delta_w_updates(log_file, plot=False):
    """
    Parses a log file of a weight-updating simulation and identifies the
    time frames when the Wang-Landau incrementor is updated.

    Parameters
    ----------
    log_file : str
        The name of the log file.
    plot : bool
        Whether to plot the Wang-Landau incrementor as a function of time.

    Returns
    -------
    t_updates : list
        A list of time frames when the Wang-Landau incrementor is updated.
    delta_updates : list
        A list of the updated Wang-Landau incrementors. Should be the same
        length as :code:`t_updates`.
    equil : bool
        Whether the weights have been equilibrated.
    """
    f = open(log_file, "r")
    lines = f.readlines()
    f.close()

    # Get the parameters
    for l in lines:  # noqa: E741
        if 'dt ' in l:
            dt = float(l.split('=')[-1])
        if 'init-wl-delta ' in l:
            init_wl_delta = float(l.split('=')[-1])
        if 'wl-scale ' in l:
            wl_scale = float(l.split('=')[-1])
        if 'weight-equil-wl-delta ' in l:
            wl_delta_cutoff = float(l.split('=')[-1])
        if 'Started mdrun' in l:
            break

    # Start parsing the data
    n = -1
    t_updates, delta_updates = [0], [init_wl_delta]
    for l in lines:  # noqa: E741
        n += 1
        if 'weights are now' in l:
            t_updates.append(int(l.split(':')[0].split('Step')[-1]) * dt / 1000)  # in ns

            # search the following 10 lines to find the Wang-Landau incrementor
            for i in range(10):
                if 'Wang-Landau incrementor is:' in lines[n + i]:
                    delta_updates.append(float(lines[n + i].split()[-1]))
                    break
        if 'Weights have equilibrated' in l:
            equil = True
            break

    if equil is True:
        delta_updates.append(delta_updates[-1] * wl_scale)

    # Plot the Wang-Landau incrementor as a function of time if requested
    # Note that between adjacen entries in t_updates, a horizontal line should be drawn.
    if plot is True:
        plt.figure()
        for i in range(len(t_updates) - 1):
            plt.plot([t_updates[i], t_updates[i + 1]], [delta_updates[i], delta_updates[i]], c='C0')
            plt.plot([t_updates[i + 1], t_updates[i + 1]], [delta_updates[i], delta_updates[i + 1]], c='C0')

        plt.text(0.65, 0.95, f'init_wl_delta: {init_wl_delta}', transform=plt.gca().transAxes)
        plt.text(0.65, 0.9, f'wl-scale: {wl_scale}', transform=plt.gca().transAxes)
        plt.text(0.65, 0.85, f'wl_delta_cutoff: {wl_delta_cutoff}', transform=plt.gca().transAxes)

        plt.xlabel('Time (ns)')
        plt.ylabel(r'Wang-Landau incrementor ($k_{B}T$)')
        plt.grid()
        plt.savefig('delta_updates.png', dpi=600)

    return t_updates, delta_updates, equil



####################################################################
#                                                                  #
#    ensemble_md,                                                  #
#    a python package for running GROMACS simulation ensembles     #
#                                                                  #
#    Written by Wei-Tse Hsu <wehs7661@colorado.edu>                #
#    Copyright (c) 2022 University of Colorado Boulder             #
#                                                                  #
####################################################################
"""
The :obj:`.utils` module provides useful utility functions.
"""
import sys
import glob
import natsort
import datetime
import subprocess
import collections
import numpy as np


class Logger:
    """
    Redirects the STDOUT and STDERR to a specified output file while preserving them on screen.

    Parameters
    ----------
    logfile : str
        Name of the output file to write the logged messages.

    Attributes
    ----------
    terminal : file object
        The file object that represents the original STDOUT (i.e., the screen).
    log : file object
        The file object that represents the logfile where messages will be written.
    """

    def __init__(self, logfile):
        """
        Initializes a Logger instance.

        Parameters
        ----------
        logfile : str
            Name of the output file to write the logged messages.
        """
        self.terminal = sys.stdout
        self.log = open(logfile, "a")

    def write(self, message):
        """
        Writes the given message to both the STDOUT and the logfile.

        Parameters
        ----------
        message : str
            The message to be written to STDOUT and logfile.
        """
        self.terminal.write(message)
        self.log.write(message)

    def flush(self):
        """
        This method is needed for Python 3 compatibility. This handles the flush command by doing nothing.
        You might want to specify some extra behavior here.
        """
        # self.terminal.log()
        pass


def run_gmx_cmd(arguments, prompt_input=None):
    """
    Runs a GROMACS command as a subprocess

    Parameters
    ----------
    arguments : list
        A list of arguments that compose of the GROMACS command to run, e.g.
        :code:`['gmx', 'mdrun', '-deffnm', 'sys']`.
    prompt_input : str or None
        The input to be passed to the GROMACS command when it prompts for input.

    Returns
    -------
    return_code : int
        The exit code of the GROMACS command. Any number other than 0 indicates an error.
    stdout : str or None
        The STDOUT of the process.
    stderr: str or None
        The STDERR or the process.

    """
    try:
        result = subprocess.run(arguments, capture_output=True, text=True, input=prompt_input, check=True)
        return_code, stdout, stderr = result.returncode, result.stdout, None
    except subprocess.CalledProcessError as e:
        return_code, stdout, stderr = e.returncode, None, e.stderr

    return return_code, stdout, stderr


def format_time(t):
    """
    Converts time in seconds to the "most readable" format.

    Parameters
    ----------
    t : float
        The time in seconds.

    Returns
    -------
    t_str : str
        A string in the format of "XX day XX hour(s) XX minute(s) XX second(s)".
    """
    hh_mm_ss = str(datetime.timedelta(seconds=t)).split(":")

    if "day" in hh_mm_ss[0]:
        # hh_mm_ss[0] will contain "day" and cannot be converted to float
        hh, mm, ss = hh_mm_ss[0], float(hh_mm_ss[1]), float(hh_mm_ss[2])
        t_str = f"{hh} hour(s) {mm:.0f} minute(s) {ss:.0f} second(s)"
    else:
        hh, mm, ss = float(hh_mm_ss[0]), float(hh_mm_ss[1]), float(hh_mm_ss[2])
        if hh == 0:
            if mm == 0:
                t_str = f"{ss:.1f} second(s)"
            else:
                t_str = f"{mm:.0f} minute(s) {ss:.0f} second(s)"
        else:
            t_str = f"{hh:.0f} hour(s) {mm:.0f} minute(s) {ss:.0f} second(s)"

    return t_str


def _autoconvert(s):
    """
    Converts input to a numerical type if possible. Used for the MDP parser.
    Modified from `utilities.py in GromacsWrapper <https://github.com/Becksteinlab/GromacsWrapper>`_.
    Copyright (c) 2009 Oliver Beckstein <orbeckst@gmail.com>

    Parameters
    ----------
    s : str or any
        The input value to be converted to a numerical type if possible. If :code:`s` is not a string,
        it is returned as is.

    Returns
    -------
    numerical : int, float, numpy.ndarray, or any
        The converted numerical value. If :code:`s` can be converted to a single numerical value,
        that value is returned as an :code:`int` or :code:`float`. If :code:`s` can be converted to
        multiple numerical values, a :code:`numpy.ndarray` containing those values is returned.
        If :code:`s` cannot be converted to a numerical value, :code:`s` is returned as is.

    Raises
    ------
    ValueError
        If :code:`s` cannot be converted to a numerical value.
    """
    if type(s) is not str:
        return s
    for converter in int, float, str:  # try them in increasing order of lenience
        try:
            s = [converter(i) for i in s.split()]
            if len(s) == 1:
                return s[0]
            else:
                return s
                """
                if len(s) != 0 and type(s[0]) == str:
                    # For the case like pull_coord1_dim = Y Y Y
                    return s
                else:
                    return np.array(s)
                """
        except (ValueError, AttributeError):
            pass
    raise ValueError("Failed to autoconvert {0!r}".format(s))


def _get_subplot_dimension(n_panels):
    """
    Gets the numbers of rows and columns in a subplot such that
    the arrangement of the .

    Parameters
    ----------
    n_panels : int
        The number of panels in the subplot.
    """
    if int(np.sqrt(n_panels) + 0.5) ** 2 == n_panels:
        # perfect square number
        n_cols = int(np.sqrt(n_panels))
    else:
        n_cols = int(np.floor(np.sqrt(n_panels))) + 1

    if n_panels % n_cols == 0:
        n_rows = int(np.floor(n_panels / n_cols))
    else:
        n_rows = int(np.floor(n_panels / n_cols)) + 1

    return n_rows, n_cols


def weighted_mean(vals, errs):
    """
    Calculates the inverse-variance-weighted mean. Note that if
    any error is 0, the simple mean will be returned.

    Parameters
    ----------
    vals : list
        A list of values to be averaged.
    errs : list
        A list of errors corresponding to the given values

    Returns
    -------
    mean : float
        The inverse-variance-weighted mean.
    err : float
        The propgated error of the mean.
    """
    if 0 in errs:
        # This could happen in the very beginning of a simulation, which could lead to an ZeroDivision warning/error.
        # In this case, we just ignore the w vector and just calculate the simple average.
        mean, err = np.mean(vals), None
        return mean, err

    w = [1 / (i ** 2) for i in errs]
    wx = [w[i] * vals[i] for i in range(len(vals))]
    mean = np.sum(wx) / np.sum(w)
    err = np.sqrt(1 / np.sum(w))

    return mean, err


def calc_rmse(data, ref):
    """
    Calculates the root mean square error (RMSE) of the given data
    with respect to the reference data.

    Parameters
    ----------
    data : list
        A list of values to be compared with the reference data.
    ref : list
        A list of reference values.

    Returns
    -------
    rmse : float
        The root mean square error.
    """
    rmse = np.sqrt(np.mean((np.array(data) - np.array(ref)) ** 2))

    return rmse


def get_time_metrics(log):
    """
    Gets the time-based metrics from a log file, including the core time (s),
    wall time, and performance (ns/day).

    Parameters
    ----------
    log : str
        The input log file.

    Returns
    -------
    t_metrics : dict
        A dictionary having following keys: :code:`t_core`, :code:`t_wall`, :code:`performance`.
    """
    t_metrics = {}
    with open(log, 'r') as f:
        # Using deque is much faster than using readlines and reversing it since we only need the last few lines.
        lines = collections.deque(f, 20)  # get the last 20 lines should be enough

    for l in lines:  # noqa: E741
        if 'Performance: ' in l:
            t_metrics['performance'] = float(l.split()[1])  # ns/day
        if 'Time: ' in l:
            t_metrics['t_core'] = float(l.split()[1])  # s
            t_metrics['t_wall'] = float(l.split()[2])  # s
            break

    return t_metrics


def analyze_REXEE_time(log_files=None):
    """
    Perform simple data analysis on the wall times and performances of all iterations of an REXEE simulation.

    Parameters
    ----------
    log_files : None or list
        A list of sorted file names of all log files.

    Returns
    -------
    t_wall_tot : float
        The total wall time GROMACS spent to finish all iterations for the REXEE simulation.
    t_sync : float
        The total time spent in synchronizing all replicas, which is the sum of the differences
        between the longest and the shortest time elapsed to finish a iteration.
    t_wall_list : list
        The list of wall times of finishing each mdrun command.
    """
    n_iter = len(glob.glob('sim_0/iteration_*'))
    if log_files is None:
        log_files = [natsort.natsorted(glob.glob(f'sim_*/iteration_{i}/*log')) for i in range(n_iter)]

    t_wall_list = []
    t_wall_tot, t_sync = 0, 0
    for i in range(n_iter):
        t_wall = [get_time_metrics(log_files[i][j])['t_wall'] for j in range(len(log_files[i]))]
        t_wall_list.append(t_wall)
        t_wall_tot += max(t_wall)
        t_sync += (max(t_wall) - min(t_wall))

    return t_wall_tot, t_sync, t_wall_list
